# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
try:
    from mpl_toolkits.basemap import Basemap
except:
    print('Plotting of maps etc. is deactivated, please install Basemap')
    
from numpy import round, log10, floor, meshgrid, arange, array, zeros, ceil,\
    log2, asarray

from matplotlib.pyplot import subplots, Polygon
import matplotlib.cm as colormaps
from random import randrange
from mpl_toolkits.mplot3d.axes3d import Axes3D
import mpl_toolkits.mplot3d.art3d as a3d
from matplotlib.pyplot import figure, draw
from os import getcwd
from os.path import join


#from geonum.base import GeoPoint, GeoVector3D
from geonum.topodata import TopoAccessError
from geonum.topodata import TopoData, TopoDataAccess
from geonum.helpers import haversine_formula, shifted_color_map, exponent

from geonum import CV2_AVAILABLE

class Map(Basemap):
    """Basemap object for drawing and plotting (on) a geographic map
    
    This object is initiated as `Basemap <http://matplotlib.org/basemap/
    users/examples.html>`_ and can therefore be used as such. 
    Added functionality mainly includes:
    
        1. Including topography data
        #. 2D and 3D plotting
        #. Handle of :class:`GeoVector3D` and :class:`GeoPoint` objects
        #. Some more high level functionality (i.e. draw text, points, 
            polygons or plot data onto map)
    
    .. note::
    
        The mapping functionality was initally developed for geographical 
        setups on local scales (i.e. several 10 km grids) including 
        handling of high resolution topography data and was not tested to 
        create large maps on global scales. In principle, this should work,
        though.
       
    """ 
    def __init__(self, *args, **kwargs):
        """Initialisation of the map object
        
        :param *args: additional non-keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        :param **kwargs: additional keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
            
        .. note::
        
            Additional possible input in **kwargs wit to ``Basemap`` 
            objects is "topo_data" which will be set in case it is a valid 
            input (i.e. :class:`TopoData`) and will be used for plotting 
            topography
        """
        super(Map, self).__init__(*args, **kwargs)

        self.topo_data = None
        #IMPORTANT HANDLES
        self.contour_lines = None
        self.contour_filled = None
        
        self.colorbars = {}
        self.points = {}
        self.lines = {}
        self.polygons = {}
        self.texts = {}
        self.plotted_data = {}
        
        self.map_scale = None
        self.meridians = None
        self.parallels = None
        
        self.default_colors = {"contour_lines"  :   "#708090",
                               "land"           :   "#FFE0B2",
                               "water"          :   "#e6e6fa"}
                               
    @property
    def color_land(self):
        """Default color of land tiles"""
        return self.default_colors["land"]

    @property
    def color_water(self):
        """Default color of ocean / sea tiles"""
        return self.default_colors["water"]
                
    def set_topo_data(self, topo):
        """Update topo data
        
        :class TopoData topo: object containing topography data
        """
        if isinstance(topo, TopoData):
            self.topo_data = topo
        
    def load_topo_data(self, mode="srtm", local_path=None):
        """Geonum wrapper for topography access"""
        ta = TopoDataAccess(mode, local_path)
        try:
            self.topo_data = ta.get_data(self.lat_ll, self.lon_ll,
                                         self.lat_tr, self.lon_tr)
            return self.topo_data
        except Exception as e:
            raise TopoAccessError(repr(e))
            
    @property
    def lon_ll(self):
        """Lower left longitude of plotted range"""
        return self.llcrnrlon
    @property
    def lat_ll(self):
        """Lower left latitude of plotted range"""
        return self.llcrnrlat
    @property
    def lon_tr(self):
        """Top right longitude of plotted range"""        
        return self.urcrnrlon
    @property
    def lat_tr(self):
        """Top right latitude of plotted range"""        
        return self.urcrnrlat
    
    @property
    def delta_lon(self):
        """Returns longitude range"""
        return abs(self.lon_tr - self.lon_ll)
        
    @property    
    def delta_lat(self):
        """Returns latitude range"""
        return abs(self.lat_tr - self.lat_ll)
        
    def fill_map(self):
        """Fill the map with default colors"""
        self.drawmapboundary(fill_color=self.color_water)
        self.fillcontinents(color=self.color_land, lake_color=self.color_water)
    
    def _prep_topo_data(self, grid_points=100):
        """Prepare topography data for map
        
        This function prepares high resolution topography for plotting, i.e.
        it reduces the resolution based on the map size
        
        :param int grid_points: number of plotted grid points (default: 100)
        
        """
        
            
        if not isinstance(self.topo_data, TopoData):
            try:
                self.load_topo_data()
            except:
                raise
    
        topo = self.topo_data
        # determine the nu
        pyr_steps = int(ceil(log2(float(len(topo.lons)) / grid_points)))
        z_max = float(topo.max)
        z_min = float(topo.min)
        z_order = floor(log10(topo.alt_range))
        X,Y = meshgrid(topo.lons, topo.lats)
        x, y = self(X, Y)
        z = topo.data
        if not CV2_AVAILABLE:
            import warnings
            warnings.warn("Could not reduce resolution of topographic data, opencv "
                          "library is not available")
        else:
            from cv2 import pyrDown
            if pyr_steps > 0:
                for k in range(pyr_steps):
                    x = pyrDown(x)
                    y = pyrDown(y)
                    z = pyrDown(z)            
        return (x, y, z, z_min, z_max, z_order)
            
    def draw_topo_contour(self, include_seabed=1, separation_levels=500):
        """Draw topography contour lines
        
         :param bool include_seabed: include seabed topography 
             (default: True)
         :param int separation_levels: separation in m of contour lines
             (default: 500)
        """

        x, y, z, z_min, z_max, z_order = self._prep_topo_data()
        if z_min > 0:
            include_seabed=1
        min_c = int(round(z_min / 10**(z_order - 1)) * 10**(z_order - 1))
        max_c = int(round(z_max / 10**(z_order - 1)) * 10**(z_order - 1))
        if include_seabed:
            levels_contour = arange(min_c, max_c, separation_levels)
        else:
            levels_contour = arange(0, max_c, separation_levels)
        
        CS1 = self.contour(x, y, z, levels_contour, linewidths=0.5,
                           colors=self.default_colors["contour_lines"])
        CS1.levels = [self._convert_float(val) for val in CS1.levels]
             
        self.ax.clabel(CS1, CS1.levels, inline=True)
        self.contour_lines = CS1   
            
    def _convert_float(self, val):
        """Custom float to string conversion for topo contour plotting
        
        :param float val: the float value to be converted into string        
        """
        s = "%.1f" %val
        if s[-1] == "0":
            return "%.0f" %val
        return "%.1f" %val
    
    def _check_ax3d(self, ax):
        """Check if input is :class:`Axes3D`"""
        if isinstance(ax, Axes3D):
            return 1
        return 0
    
    def set_ticks_topo_colorbar(self, start, stop, step):
        """Update ticks of topo colorbar"""
        cb = self.colorbars["topo"]
        ticks = arange(start, stop, step)
        cb.set_ticks(ticks)
        draw()
        
    def draw_topo(self, insert_colorbar=False, include_seabed=True,
                    max_grid_points=500, cmap_div=colormaps.coolwarm,
                    cmap_seq=colormaps.Oranges, alpha=0.5, ax=None):
        """Draw topography into map
        
        :param bool insert_colorbar: draws a colorbar for altitude
            range (default: False)
        :param bool include_seabed: include seabed topography 
            (default: True)
        :param int max_grid_points: resolution of displayed topo data 
            points (makes it faster in interactive mode, default: 500)
        :param str cmap_div: name of a diverging colormap (this one is 
            used if :arg:`include_seabed` is True, and the cmap is shifted 
            such , that white colors correspond to sea level altitude, 
            default: "coolwarm")
        :param str cmap_seq: name of a sequential colormap (this one is 
            used if :arg:`include_seabed` is False, default: "Oranges")
        :param float alpha: Alpha value (transparency) of plotted 
            topography
        :param ax: matplotlib axes object
        """
        try:  
            if ax is None:
                ax = self.ax
            if ax is None:
                fig, ax = subplots(1, 1, figsize=(16,10))
                self.ax = ax
    
            x, y, z, z_min, z_max, z_order =\
                self._prep_topo_data(grid_points=max_grid_points)
                
            
            delz = z_max - z_min #in m
            exp = exponent(delz) - 1
            start_alt = round((z_min - delz*0.1)*10**(-exp))*10**exp
            stop_alt = round((z_max + delz*0.1)*10**(-exp))*10**exp
            if z_min > 0:
                include_seabed = 1
                
            if not include_seabed:
                start_alt = 0
            
            z_step = (stop_alt - start_alt) / 1000. 
            levels_filled = arange(start_alt, stop_alt, z_step)
            
            if levels_filled[0] < 0:          
                shifted_cmap = shifted_color_map(z_min, z_max, cmap_div)
                
                cs2 = ax.contourf(x, y, z, levels_filled, cmap=shifted_cmap,
                                  extend="both", alpha=alpha)
                                  
            elif levels_filled[0] >= 0:        
                cs2 = ax.contourf(x, y, z, levels_filled, cmap=cmap_seq,
                                  alpha=1.0, extend="min")
                self.contour_filled = cs2
                    
            if insert_colorbar:              
                self.insert_colorbar("topo", cs2, label="Altitude [m]")
                exp = exponent(delz)
                step = round(delz*10**(-exp))*10**(exp-1)
                self.set_ticks_topo_colorbar(start_alt, stop_alt, step)
        except Exception as e:
            raise
            msg=("Could not draw topography in high res, using default "
                 "etopo() instead...")
            print(msg + repr(e))
            self.etopo()
            
    def draw_topo_old(self, insert_colorbar=False, include_seabed=True,
                    max_grid_points=500, cmap_div=colormaps.coolwarm,
                    cmap_seq=colormaps.Oranges, alpha=0.5, ax=None):
        """Draw topography into map
        
        :param bool insert_colorbar: draws a colorbar for altitude
            range (default: False)
        :param bool include_seabed: include seabed topography 
            (default: True)
        :param int max_grid_points: resolution of displayed topo data 
            points (makes it faster in interactive mode, default: 500)
        :param str cmap_div: name of a diverging colormap (this one is 
            used if :arg:`include_seabed` is True, and the cmap is shifted 
            such , that white colors correspond to sea level altitude, 
            default: "coolwarm")
        :param str cmap_seq: name of a sequential colormap (this one is 
            used if :arg:`include_seabed` is False, default: "Oranges")
        :param float alpha: Alpha value (transparency) of plotted 
            topography
        :param ax: matplotlib axes object
        """
        try:  
            if ax is None:
                ax = self.ax
            if ax is None:
                fig, ax = subplots(1, 1, figsize=(16,10))
                self.ax = ax
    
            x, y, z, z_min, z_max, z_order =\
                self._prep_topo_data(grid_points=max_grid_points)
                
            if z_min > 0:
                include_seabed = 1
                
            z_step = (z_max - z_min) / 1000. 
            
            if include_seabed:
                levels_filled = arange(z_min, z_max + z_step, z_step)
            else:
                levels_filled = arange(0, z_max + 1, z_step)
            if levels_filled[0] < 0:          
                shifted_cmap = shifted_color_map(z_min, z_max, cmap_div)
                
                cs2 = ax.contourf(x, y, z, levels_filled, cmap=shifted_cmap,
                                  extend="both", alpha=alpha)
                                  
            elif levels_filled[0] >= 0:        
                print("HEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEERE")
                cs2 = ax.contourf(x, y, z, levels_filled, cmap=cmap_seq,
                                  alpha=1.0, extend="min")
                self.contour_filled = cs2
                    
            if insert_colorbar:              
                self.insert_colorbar("topo", cs2, label="Altitude [m]")
        
        except Exception as e:
            raise
            msg=("Could not draw topography in high res, using default "
                 "etopo() instead...")
            print(msg + repr(e))
            self.etopo()
         
    def draw_topo_3d(self, num_ticks=4, cmap="Oranges", alpha=0.5,
                     contour_color="#708090", contour_antialiased=True, 
                     contour_lw=0.2, ax=None):
        """Draw topography into 3D axis
        
        Parameters
        ----------        
        num_ticks : int 
            number of lon / lat axis ticks (default: 4)
        cmap : str
            string ID of the colormap used (default: "Oranges")
        alpha : flaot
            Alpha value (transparency) of plotted topography (default: 0.5)
        contour_color : str
            string specifying color of contour lines colors of contour lines 
            (default: "#708090")
        contour_antialiased : bool
            apply antialiasing to surface plot of topography, defaults to False
        contour_lw : 
            width of drawn contour lines, defaults to 0.5, use 0 if you do not 
            want contour lines inserted
        ax : 
            axes object
        """
        if not self._check_ax3d(ax):
            if self._check_ax3d(self.ax):
                ax = self.ax
            else:
                fig = figure()
                ax = Axes3D(fig)
            
        x, y, z, z_min, z_max, z_order = self._prep_topo_data()
        tickformatter = "{:.2f}"    
        ax.plot_surface(x, y, z, rstride=1, cstride=1, cmap=cmap, alpha=alpha, 
                        linewidth=contour_lw, antialiased=contour_antialiased, 
                        edgecolors=contour_color, vmin=z_min, vmax=z_max, 
                        zorder=1)
        #xlabels=ax.get_xticklabels()
        ax.set_xlabel("Lons")
        ax.set_ylabel("Lats")
        ax.set_zlabel("Altitude [m]")
        lon_step, lat_step = self.delta_lon * .33, self.delta_lat * .33
        lon_tick_array, lat_tick_array=[], []
        for k in range(num_ticks):
            lon_tick_array.append(self.lon_ll + lon_step * k)
            lat_tick_array.append(self.lat_ll + lat_step * k)
        #lon_tick_array, lat_tick_array = self._prep_coord_ticks()
        lon_coords=[]
        lon_labels=[]
        for lon in lon_tick_array:
            x,_=self(lon,self.lat_ll)
            lon_coords.append(x)
            lon_labels.append(tickformatter.format(lon))
        
        lat_coords=[]
        lat_labels=[]
        for lat in lat_tick_array:
            _, y = self(self.lon_ll, lat)
            lat_coords.append(y)
            lat_labels.append(tickformatter.format(lat))
            
        ax.set_yticks(lat_coords)
        ax.set_yticklabels(lat_labels)
        ax.set_xticks(lon_coords)
        ax.set_xticklabels(lon_labels)
        self.ax = ax
        
    def draw_mapscale_auto(self, **kwargs):
        """Insert a map scale 
        
        Determines missing input parameters automatically and calls 
        :func:`drawmapscale` 
        
        :param kwargs: key word arguments for :func:`drawmapscale` (missing
            ones will be set automatically)
        """
        l = self._len_diag()  
        exp = exponent(l)
        l = round(l * 10**(-exp))*10**(exp)/5.0
        lat_center, lon_center = self.get_map_center()
        str_format = '%d'
        if not "length" in kwargs:
            kwargs["length"] = l
        if l < 40:
            str_format = '%.1f'
# =============================================================================
#         if not "fontsize" in kwargs:
#             kwargs["fontsize"] = 8
# =============================================================================
        if not "lon0" in kwargs:
            kwargs["lon0"] = lon_center
            kwargs["lat0"] = lat_center
        if not "lon" in kwargs:
            kwargs["lon"] = lon_center - self.delta_lon * 0.3
            kwargs["lat"] = lat_center - self.delta_lat * 0.4
        if not "barstyle" in kwargs:
            kwargs["barstyle"] = "fancy"
        if not "units" in kwargs:
            kwargs["units"] = "km"
        if not "format" in kwargs:
            kwargs["format"] = str_format
        self.map_scale = self.drawmapscale(**kwargs)
        
    def remove_map_scale(self):
        """Remove scale"""
        if self.map_scale is not None:
            for i in self.map_scale:
                try:
                    i.remove()
                except:
                    print(("Warning in remove_map_scale: item %s could not be "
                        "removed" %i))
        self.fig.canvas.draw()
    
    def _prep_coord_ticks(self, lon_tick=None, lat_tick=None):
        """Prepare coordinate ticks for parallels and meridians
        
        :param float lon_tick: longitude separation (in decimal degrees).
            Will be set automatically if unspecified
        :param float lat_tick: latitude separation (in decimal degrees)
            Will be set automatically if unspecified
        """
        if lon_tick is None:
            pot_lon = floor(log10(self.delta_lon))
            lon_tick = floor(self.delta_lon / 10**pot_lon) * 10**pot_lon / 4
        if lat_tick is None:
            pot_lat = floor(log10(self.delta_lat))
            lat_tick = floor(self.delta_lat / 10**pot_lat) * 10**pot_lat / 3
            
        lon_tick_array = arange(lon_tick * int((self.lon_ll - self.delta_lon 
                                * 0.3) / lon_tick), lon_tick * int((self.lon_tr 
                                + self.delta_lon * 0.3) / lon_tick), lon_tick)
                                
        lat_tick_array = arange(lat_tick * int((self.lat_ll - self.delta_lat 
                                * 0.3) / lat_tick), lat_tick * int((self.lat_tr 
                                + self.delta_lat * 0.3) / lat_tick), lat_tick)
                                
        return lon_tick_array, lat_tick_array
    
    def _len_diag(self):
        """Returns the lenght from LL point to TR point in km"""
        return self.haversine(self.lon_ll,self.lat_ll,self.lon_tr,self.lat_tr)
    
        
    def draw_coordinates(self, lat_tick=None, lon_tick=None,
                         labelslon=[0,0,0,1], labelslat=[1,0,0,0], 
                         **kwargs):
        """Draws meridians and parallels 
        
        :param float lat_tick: latitude separation (in decimal degrees)
        :param float lon_tick: longitude separation (in decimal degrees)
        :param labelslon: see basemap docs (default: [0,0,0,1])
        :param labelslat: see basemap docs (default: [1,0,0,0])
        :param **kwargs: additional keyword arguments (passed to 
            :func:`drawmeridians` and :func:`drawparallels`)
        """
        if not "color" in kwargs:
            kwargs["color"] = "gray"
        if not "fmt" in kwargs:
            digs = '%d' %(4 - int(floor(log10(self._len_diag()))))
            kwargs["fmt"] = "%." + digs + "f"
        lon_tick_array, lat_tick_array = self._prep_coord_ticks(lon_tick,
                                                                lat_tick)
        meridians = self.drawmeridians(lon_tick_array, labels=labelslon,
                                       **kwargs)
        for m in meridians:
            try:
                meridians[m][1][0].set_rotation(25)
            except:
                pass
        self.meridians = meridians
        self.parallels = ps =self.drawparallels(lat_tick_array, 
                                                labels=labelslat, 
                                                **kwargs)
        for p in ps:
            try:
                ps[p][1][0].set_rotation(25)
            except:
                print("Could not rotate parallel...")
            
    def remove_coordinates(self):
        """Remove drawn parallels and meridians"""
        if self.meridians is not None:
            for Dict in [self.meridians, self.parallels]:
                for key in Dict:
                    for item in Dict[key]:
                        try:
                            item[0].remove()
                        except:
                            print("Object %s could not be removed" %item)
        self.fig.canvas.draw()
    
    def legend(self, ax=None, **kwargs):
        """Insert a legend"""
        if ax is None:
            ax = self.ax
# =============================================================================
#         if not "fontsize" in kwargs:
#             kwargs["fontsize"] = 10
# =============================================================================
        ax.legend(loc="best", fancybox=True, framealpha=0.7, **kwargs)
                                                                
    def draw_geo_point_2d(self, p, add_name=False, ax=None, **kwargs):
        """Draw a GeoPoint into 2D basemap
        
        :param GeoPoint p: the actual point        
        :param **kwargs: passed to matplotlib plotting function
            
        """
        if ax is None:
            ax = self.ax
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not "c" in kwargs:
            kwargs["c"]="lime"
        x, y = self(p.longitude, p.latitude) #maps the geo coordinates to figure coordinates
                
        handle = ax.plot(x, y, **kwargs) 
        self.points[p.name] = handle
        if add_name:
            self.write_point_name_2d(p)
        return handle
    
    def draw_geo_vector_2d(self, vec, **kwargs):
        """Draw a :class:`GeoVector3D` into 2D map
        
        :param GeoVector3D vec: the vector 
        
        .. note::
        
            Anchor must be set in the :class:`GeoVector3D` object  
        """
        if not vec.type() == "GeoVector3D":
            raise AttributeError("Wrong input, need :class:`GeoVector3D` "
                                                                "object")
        elif not vec.anchor.type() == "GeoPoint":
            raise AttributeError("Vector anchor not set or wrong type..")

        if not any([x in kwargs for x in["ls", "linestyle"]]):
            kwargs["ls"] = "--"
            
        a = vec.anchor
        pf = a + vec
        x0, y0 = self(a.longitude, a.latitude)
        x1, y1 = self(pf.longitude, pf.latitude)
        return self.draw_line_2d(vec.name, a.latitude, a.longitude,
                                 pf.latitude, pf.longitude, **kwargs)
    
    def add_geo_points_3d(self, pts, distances=None, marker="x", color="b", 
                          connect=True, connect_style="--", alt_offset_m=0):
        """Draws a list of :class:`GeoPoint` objects into the map
        
        :param list pts: geopoint objects
        :param color: color of points (passed to plot function, default: "b")
        :param marker: marker of points (default: "x")
        :param connect: if True, points the points are connected with each 
            other
        :param str connect_style: line style of connection
        """
        xs, ys, zs = [], [], []
        for p in pts:
            try:
                px, py= self(p.lon.decimal_degree,p.lat.decimal_degree)
                xs.append(px), ys.append(py), zs.append(p.altitude)
                self.draw_geo_point_3d(p, alt_offset_m=alt_offset_m, 
                                       marker=marker, s=20, c=color)
            except:
                print("Failed to add %s to map" %p)
        if distances is not None and len(distances) == len(pts):
            sc = self.ax.scatter(xs, ys, asarray(zs)+alt_offset_m, zdir='z', 
                                 s=20, c=distances, cmap="Oranges", 
                                 zorder=100000)
            zlabel = "Distance [%km]"
            self.insert_colorbar("dists", sc, label=zlabel)
        elif connect:
            self.ax.plot(xs, ys, asarray(zs)+alt_offset_m, ls=connect_style, 
                         c=color, lw=2, zorder=100000)   
                                                
    def draw_geo_point_3d(self, p, ax=None, alt_offset_m=0.0, **kwargs):
        """Draw a GeoPoint into 3D basemap
        
        :param GeoPoint p: the actual point        
        :param ax: matplotlib axes object (3d axis)
        :param float alt_offset_m: plot point with altitude offset, default 0
        :param **kwargs: passed to matplotlib plotting function
        
        .. note::
        
            The basemap needs to be set up with Axes3d object
            
        """
        if not self._check_ax3d(ax):
            if not self._check_ax3d(self.ax):    
                raise TypeError("No 3D Axes object available...")
            ax=self.ax
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"

        if not isinstance(self.ax, Axes3D):
            raise ValueError("Need :class:`Axes3D` object as input...")
        x0, y0 = self(p.longitude, p.latitude)
        z0 = p.altitude + alt_offset_m#*1000
        handle = ax.scatter(x0, y0, z0, zorder=100000, **kwargs)
        self.points[p.name] = handle
        return handle
                                 
    def draw_geo_vector_3d(self, vec, ax=None,  **kwargs):
        """Draw a :class:`GeoVector3D` into 3D map
        
        :param GeoVector3D vec: the vector 
        
        .. note::
        
            Anchor must be set in the :class:`GeoVector3D` object  
            
        """ 
        if ax is None:
            ax = self.ax
        try:
            if not isinstance(ax, Axes3D):
                raise ValueError("Need :class:`Axes3D` object as input...")
            elif not vec.type() == "GeoVector3D":
                raise AttributeError("Wrong input, need :class:`GeoVector3D` "
                                                                    "object")
            elif not vec.anchor.type() == "GeoPoint":
                raise AttributeError("Vector anchor not set or wrong type..")
    
            if not any([x in kwargs for x in["ls", "linestyle"]]):
                kwargs["ls"] = "--"
            
            a = vec.anchor
            pf = a + vec
            x0, y0 = self(a.longitude, a.latitude)
            z0 = a.altitude#*1000
            x1, y1 = self(pf.longitude, pf.latitude)
            z1 = pf.altitude#*1000
    
            l = a3d.Line3D((x0,x1), (y0,y1), (z0,z1), **kwargs)
            handle = ax.add_line(l)
            self.lines[vec.name] = handle
            return pf

        except:
            print(vec.anchor.type())
            raise
    
    def add_polygon_2d(self, points=[], poly_id="undefined", ax=None, **kwargs):
        """Add a polygon specified by list of input points
        
        :param list points: list with :class:`GeoPoint` objects
        :param str poly_id: string ID of this object (e.g. for 
            deletion, default: "undefined")
        """
        if ax is None:
            ax = self.ax
        if not "label" in kwargs:
            kwargs["label"] = poly_id
        coords=[]
        for p in points:
            try:
                coords.append(self(p.longitude, p.latitude))
            except Exception as e:
                print("Failed to add one point to poly: " + repr(e))
        polygon = Polygon(coords, **kwargs)
        ax.add_patch(polygon)
        
    def add_polygon_3d(self, points=[], poly_id="undefined", ax=None, **kwargs):
        """Add a polygon specified by list of input points 
        
        :param list points: list with :class:`GeoPoint` objects
        :param str poly_id: string ID of this object (e.g. for 
            deletion, default: "undefined")
        """
        if ax is None:
            ax = self.ax
        if not "label" in kwargs:
            kwargs["label"] = poly_id
        xs, ys, zs = [], [], []
        for p in points:
            x,y=self(p.longitude, p.latitude)
            xs.append(x)
            ys.append(y)
            zs.append(p.altitude)#*1000)
        coords = [list(zip(xs, ys, zs))]
        poly_coll = a3d.Poly3DCollection(coords, **kwargs)
        ax.add_collection3d(poly_coll)
           
    def draw_line_2d(self,line_id, lat0, lon0, lat1, lon1, **kwargs):
        """Draw a line between 2 geo coordinates
        
        :param str line_id: ID of the line artist (for deletion management)
        :param float lon0: start longitude 
        :param float lat0: start latitude 
        :param float lon1: stop longitude 
        :param float lat1: stop latitude
        :returns: line object
        """
        line = self.drawgreatcircle(lon0, lat0, lon1, lat1, **kwargs)
        self.lines[line_id] = line
        return line
    
    def write_point_name_2d(self, p, dist=0, angle=0, ax=None, **kwargs):
        """Annotate name of point to point in map
        
        :param GeoPoint p: the point
        :param float dist: distance of annotation in km (default: 0)
        :param float angle: angular direction of  diplacement (default: 0)
        :param ax: matplotlib axes instance
        """
        if dist == 0:
            self.draw_text_2d(p.longitude, p.latitude, p.name, ax, **kwargs) 
        else:
            pt = p.offset(azimuth=angle, dist_hor=dist)
            self.annotate_text_2d(pt.longitude, pt.latitude, p.name,
                                  p.longitude, p.latitude, ax, **kwargs)
    
    def annotate_text_2d(self, lon_text, lat_text, text, lon_point, lat_point,
                         ax=None, **kwargs):
        """Annotate text to a certain position in a 2D basemap
        
        :param float lon_text: longitude coordinate of text position
        :param float lat_text: latitude coordinate of text position
        :param str text: the actual text
        :param float lon_point: longitude of text annotation
        :param float lat_point: latitude of text annotation
        :param **kwargs: additional keyword arguments passed to 
            :func:`annotate`
        """
        if ax is None:
            ax = self.ax
        x_text, y_text = self(lon_text, lat_text)  
        x_point, y_point = self(lon_point, lat_point)
        t = ax.annotate(text, xy=(x_point, y_point), xytext=(x_text, y_text), 
                        arrowprops=dict(arrowstyle="->", connectionstyle=
                        "arc,angleA=10,armA=20,rad=6", shrinkA=2.0, 
                        shrinkB=10), **kwargs)
                        
        self.texts[text] = t
        return t
            
    def draw_text_2d(self, lon, lat, text, ax=None, **kwargs):
        """Draw a text into a 2D map
                
        :param float lon: longitude of text
        :param float lat: latitude of text
        :param str text: the actual text
        :param **kwargs: draw parameters 
        """
        if ax is None:
            ax = self.ax
        x, y = self(lon, lat)  
        t = ax.text(x, y, text, **kwargs)
        self.texts[text] = t
        return t
    
    def draw_text_3d(self, lon, lat, alt, text, ax=None, **kwargs):
        """Draw a text into a 3D map
        
        :param float lon: longitude of text
        :param float lat: latitude of text
        :param float alt: altitude of text (in m)
        :param str text: the actual text
        :param **kwargs: draw parameters 
        """
        if not self._check_ax3d(ax):
            if not self._check_ax3d(self.ax):    
                raise TypeError("No 3D Axes object available...")
            ax = self.ax
        x, y = self(lon, lat)  
        t = ax.text(x, y, alt, text, zorder=1e6, **kwargs)
        self.texts[text] = t
        return t
    
    def draw_data_scatter_2d(self, data=None, lons=None, lats=None,
                             unit=None, data_id="n/d", ax=None, **kwargs):
        """Draw data into map
        
        If input is unspecified, a random test data set will be created and
        plotted onto the map.
        
        :param float data: the actual data
        :param float lons: array with longitude coordinates
        :param float lats: array with latitude coordinates
        :param str unit: Unit of Data for labelling
        :param str data_id: string ID under which data will be stored
        :param ax: axes object
        :param **kwargs: keyword arguments passed to :func:`scatter`
        
        """
        if ax is None:
            ax = self.ax
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not "s" in kwargs:
            kwargs["s"] = 30
        if not any([x in kwargs for x in ["edgecolor", "edgecolors"]]):
            kwargs["edgecolor"] = "none"
        
        if data is None:
            print("No input data, create random data")
            lons, lats, data = self.make_random_data()
            data_id = "TestData"
            unit = 'n/d'
    
        x, y = self(lons, lats)
        z = array(data)
        sc = ax.scatter(x, y, c=z, **kwargs)
        zlabel = "%s [%s]" %(data_id, unit)
        self.insert_colorbar(data_id, sc, label=zlabel)
        self.plotted_data[data_id] = sc

    def insert_colorbar(self, obj_id, obj, label=None, ax=None, **kwargs):
        """Insert a colorbar into the map
        
        :param obj_id: save ID
        :param obj: the actual artist to which the colorbar belongs
        :param str label: colorbar label
        :param **kwargs: further arguments for :func:`colorbar` call 
            (see matplotlib docs)
        """
        if ax is None:
            ax = self.ax
        cb = self.fig.colorbar(obj, shrink=0.9, ax=ax, **kwargs)
        if isinstance(label, str):
            cb.set_label(label, fontsize=16)
        self.colorbars[obj_id] = cb
        ax.figure.canvas.draw()
        
    def remove_colorbar(self, cb_id):
        """Remove a drawn colorbar
        
        :param str cb_id: string ID of colorbar        
        """
        if not cb_id in self.colorbars:
            raise KeyError("No colorbar found with ID: " + cb_id)
        cb = self.colorbars[cb_id]
        cb.remove()
        del self.colorbars[cb_id]
        self.fig.canvas.draw()
    
    @property
    def set_ax(self):
        """Check current axes and write into ``self.ax``"""
        self.ax = self._check_ax()
        return self.ax  
        
    @property
    def fig(self):
        """The current figure canvas"""
        return self.set_ax.figure
        
    def remove_all_colorbars(self):
        """Remove all colorbars from map"""
        cbs = self.colorbars
        for key, cb in cbs.items():
            cb.remove()
            del self.colorbars[key]
        self.colorbars = {}
        self.fig.canvas.draw()
        
    """Calculations etc
    """    
    
    def haversine(self, lon0, lat0, lon1, lat1):
        """Haversine formula (Distance between two geo coordinates)
        
        Approximate horizontal distance between 2 points assuming a spherical 
        earth
        
        :param float lon0: longitude of first point in decimal degrees
        :param float lat0: latitude of first point in decimal degrees
        :param float lon1: longitude of second point in decimal degrees
        :param float lat1: latitude of second point in decimal degrees
        """
        return haversine_formula(lon0, lat0, lon1, lat1)
            
    def get_map_center(self):
        """Returns center coordinates of map (lat, lon)"""
        lon_center = self.lon_ll + self.delta_lon / 2.0
        lat_center = self.lat_ll + self.delta_lat / 2.0
        return (lat_center, lon_center)
        
    """Other stuff"""
    def make_random_data(self,total_number=200, value_range=[-100, 100, 1]):
        """Create random data in the current regime"""
        lons, lats = zeros(total_number), zeros(total_number)
        data = zeros(total_number)
        for k in range(total_number):
            lons[k] = 0.1 * randrange(self.lon_ll * 10, self.lon_tr * 10, 1)
            lats[k] = 0.1 * randrange(self.lat_ll * 10, self.lat_tr * 10, 1)
            data[k] = randrange(value_range[0], value_range[1], value_range[2])
            
        
        return lons, lats, data
    
    """I/O, printing, etc...
    """
    def print_map_info(self):
        """Print some basic info about this map"""
        lat_center,lon_center=self.get_map_center()
        print()
        print("-------------------------------")
        print("------- MAP INFORMATION -------")
        print("-------------------------------")
        print()
        print("(Lon|Lat) lower left: (%s|%s)" %(self.lon_ll, self.lat_ll))
        print("(Lon|Lat) top right: (%s|%s)" %(self.lon_tr, self.lat_tr))
        print("(Lon|Lat) center: (%s|%s)" %(lon_center, lat_center))
        print("Projection: %s" %self.projection)
        
    def save_as(self, save_dir=None, save_name=None, ftype="png"):
        """Save current figure as image"""
        if save_dir is None:
            save_dir = getcwd()
        if save_name is None:
            save_name = "geonum_map"
        self.fig.savefig(join(save_dir, save_name + ftype))
        
if __name__ == "__main__":
    from matplotlib.pyplot import close
    from geonum import GeoSetup
    close("all")
    s = GeoSetup()
    s.create_test_data()
    s.plot_2d()
    


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the GeoSetup class, a high level object for managing 
muliple geo point and geo vector objects.
"""
from geonum import BASEMAP_AVAILABLE
if BASEMAP_AVAILABLE:
    from geonum.mapping import Map
from numpy import asarray, nanmin, nanmax
from os.path import exists
from traceback import print_exc
from warnings import warn
from matplotlib.pyplot import get_cmap, figure

from geonum.base import GeoPoint, GeoVector3D
from geonum.topodata import TopoDataAccess, TopoData

class GeoSetup(object):
    """The GeoSetup class represents acollection of GeoPoints and vectors
    """
    def __init__(self, points=[], vectors=[], lat_ll=None, lon_ll=None, 
                 lat_tr=None, lon_tr=None, id="MyGeoSetup", 
                 topo_access_mode="srtm", local_topo_path=None, 
                 cmap_vecs="Greens"):
        """Init object
        
        Parameters
        ----------
        points : list
            list of :class:`GeoPoint` objects to be included in this setup
        vectors : list
            list of :class:`GeoVector3D` objects to be included in this\
            setup
        lat_ll : :obj:`float`, optional
            lower left latitude of regime
        lon_ll : :obj:`float`, optional
            lower left longitude of regime
        lat_tr : :obj:`float`, optional
            top right latitude of regime
        lon_tr : :obj:`float`, optional
            top right longitude of regime
        id : str 
            identification string of this setup
        topo_access_mode : str
            topo data mode, default is SRTM
            (see :class:`TopoDataAccess` for details)
        local_topo_path : str 
            local path were topography data (e.g. ETOPO1 data) is stored
        cmap_vecs : str
            String specifying a valid matplotlib colormap supposed to be
            used for drawing :class:`GeoVector3D` objects into overview
            maps
        """
        self.id = id
        self.points = {}
        self.vectors = {}
        
        try:
            cmap = get_cmap(cmap_vecs)
        except:
            cmap = get_cmap("Greens")
        self.cmap = cmap
        
        self.topo_access = TopoDataAccess(topo_access_mode, 
                                          local_topo_path) 
        self.topo_data = None
        try:
            iter(points)
        except TypeError:
            if isinstance(points, GeoPoint):
                points = [points]
            else:
                print("Invalid input points: ", points) 
                points = []
        try:
            iter(vectors)
        except:
            if isinstance(points, GeoPoint):
                points = [points]
            else:
                print("invalid input vectors: ", vectors)
                vectors = []
                
        for pt in points:
            if isinstance(pt, GeoPoint):
                self.add_geo_point(pt)
        for vec in vectors:
            if isinstance(vec, GeoVector3D):
                self.add_geo_vector(vec)
                
        self.set_borders_from_points()
        #temporary creation of points ll and tr in case there is some valid input
        try:
            self.new_geo_point(lat_ll, lon_ll, name="ll")
            self.new_geo_point(lat_tr, lon_tr, name="tr")
        except (TypeError, ValueError):
            pass
      
    def has_points(self):
        """Returns True, if this setup includes GeoPoints, False if not"""
        if not bool(self.points):
            return False
        return True
    
    def create_test_data(self):
        """Create exemplary test data set"""
        source = GeoPoint(37.751005,  14.993435, name="Etna")
        instrument = GeoPoint(37.765755,  15.016696, name="Observatory")
        self.add_geo_points(source, instrument)
        self.set_borders_from_points()
        plume = GeoVector3D(azimuth=83, dist_hor = self.magnitude,
                            elevation=0, anchor=source, name="plume")
        view_dir = GeoVector3D(azimuth=160, dist_hor=self.magnitude,
                               elevation=8, anchor=instrument, name="cfov")
                                
        self.add_geo_vectors(plume, view_dir)
        
    def set_local_topo_path(self, p):
        """Sets local path for Etopo1 data files can be found
        
        :param str p: path 
        
        .. note::
        
            The default topomode is "srtm" which provides online access, so
            it is not mandatory to provide topography data locally. However, 
            the SRTM model has no global coverage, so there might be need to 
            use another of the provided topomodes and provide the respective
            files locally.
            
        """
        if p is None:
            return
        if not exists(p):
            raise IOError("Path %s could not be set" %p)
        self.topo_access.local_path = p
    
    def change_topo_mode(self, new_mode="srtm", local_path=None):
        """Change the current mode for topography data retrieval and optionally
        change the local topo data path at the same time
        
        :param str new_mode: the new mode 
        :param local_path (None): if not None and valid, update local topo access
            in ``self.topo_access``
        """
        self.topo_access.set_mode(new_mode, local_path)
    
    def get_topo(self):
        """Get current topo data"""
        if not isinstance(self.topo_data, TopoData):
            self.load_topo_data()
        return self.topo_data
        
    def load_topo_data(self):
        """Load topography data 
        
        .. note:: 
        
            The loaded :class:`TopoData` object will also be set in all 
            :class:`GeoPoint` objects belonging to this setup
            
        """
        if "ll" not in self.points:
            self.set_borders_from_points()
        self.topo_data = self.topo_access.get_data(self.ll.latitude,
                                                   self.ll.longitude,
                                                   self.tr.latitude,
                                                   self.tr.longitude)
        for p in list(self.points.values()):
            p.set_topo_data(self.topo_data)
        
    @property
    def ll(self):
        """Return lower left point of topo data range """
        try:
            return self.points["ll"]
        except AttributeError:
            print("Lower left corner (GeoPoint) not yet defined in GeoSetup")
    
    @ll.setter
    def ll(self, value):
        if not isinstance(value, GeoPoint):
            raise TypeError("Could not set lower left coordinate in "
                "GeoSetup: need GeoPoint object")
        self.points["ll"] = value
        
    @property
    def tr(self):
        """Return lower left point of topo data range"""
        try:
            return self.points["tr"]
        except AttributeError:
            print("Top right corner (GeoPoint) not yet defined in GeoSetup")
            pass
    
    @tr.setter
    def tr(self, value):
        if not isinstance(value, GeoPoint):
            raise TypeError("Could not set top right coordinate in "
                            "GeoSetup: need GeoPoint object")
        self.points["tr"] = value
        
    @property
    def lon_ll(self):
        """Lower left corner of object regime"""
        return self.ll.longitude
    
    @property
    def lat_ll(self):
        """Lower left corner of object regime"""
        return self.ll.latitude
        
    @property
    def lon_tr(self):
        """Lower left corner of object regime"""
        return self.tr.longitude
    
    @property
    def lat_tr(self):
        """Lower left corner of object regime"""
        return self.tr.latitude
    
    @property
    def delta_lon(self):
        """Returns longitude range"""
        return abs(self.lon_tr - self.lon_ll)
        
    @property    
    def delta_lat(self):
        """Returns latitude range"""
        return abs(self.lat_tr - self.lat_ll)
        
    @property
    def center_coordinates(self):
        """Lat / Lon coordinates of center of data"""
        return (self.lat_ll + self.delta_lat / 2., 
                self.lon_ll + self.delta_lon / 2.)
        
    def add_geo_points(self, *args):
        """Add multiple GeoPoints to the collection
        
        :param *args: arbitrary amount of new geo points        
        """
        for arg in args:
            self.add_geo_point(arg)
    
    def has_point(self, name):
        """Checks if point with input name exists
        
        :param str key: name of point
        :return: bool
        """
        if name in self.points:
            return True
        return False
    
    def has_vector(self, name):
        """Checks if vector with input name exists"""
        if name in self.vectors:
            return True
        return False
        
    def add_geo_point(self, pt):
        """Add :class:`GeoPoint` to this collection
        
        :param GeoPoint pt: the new point
        """
        try:
            if pt.name in self.points:
                print(("Point ID %s already exists in GeoSetup" %(pt.name)))
                pt2 = self.points[pt.name]
                if pt.almost_equal(pt2) and pt.altitude == pt2.altitude:
                    print("Point is unchanged")
                    return
                print("Updating name of existing GeoPoint to: %s_old" %pt.name)
                pt2.name = pt2.name + "_old"
                self.points[pt2.name] = pt2
            self.points[pt.name] = pt
            if not isinstance(pt.topo_data, TopoData):
                pt.set_topo_data(self.topo_data)
        
        except Exception as e:
            print("Geopoint could not be added: " + repr(e))
            
    def set_geo_point(self, p_id, pt):
        """Update an existing GeoPoint in the collection
        
        :param str id: id of existing point in ``self.points``
        :param GeoPoint pt: a new geo_point
        """
        if not isinstance(pt, GeoPoint):
            raise TypeError("Wrong input: " + type(pt))
        self.points[p_id] = pt
        
    def add_geo_vectors(self, *args):
        """Add multiple GeoPoints to the collection"""
        for arg in args:
            self.add_geo_vector(arg)
            
    def add_geo_vector(self, vec):
        """Add :class:`GeoVector3D` to this collection
        
        :param GeoVector3D vec: should be clear ;)
        """
        if not isinstance(vec, GeoVector3D):
            print(("Error adding GeoVector3D, wrong input type, need "
                  ":class:`GeoVector3D` object, input type: %s" %type(vec)))
            return
        if vec.name in self.vectors:
            print(("Vector ID %s already exists in %s" %(vec.name, self)))
            vec2 = self.vectors[vec.name]
            if (vec2.almost_equals(vec) and vec2.dz == vec.dz 
                and vec.anchor == vec2.anchor):
                print("Vector is unchanged")
                return 
            print("Updating name of existing vector to: %s_old" %vec.name)
            vec2.name = vec2.name + "_old"
            self.vectors[vec2.name] = vec2
        self.vectors[vec.name] = vec
    
    def delete_geo_point(self, name):
        """Remove one of the geo_points from the collection
        
        :param str name: name of geo point        
        """
        del self.points[name]
        
    
    def delete_geo_vector(self, name):
        """Remove one of the vectors from the collection
        
        :param str name: name of geo vector
        """
        del self.vectors[name]
            
    def new_geo_point(self, *args, **kwargs):
        """Create new geo_point and add to collection
        
        :param **kwargs: see :class:`GeoPoint` for initiation info
        """
        try:
            self.add_geo_point(GeoPoint(*args, **kwargs))
        except (TypeError, ValueError):
            return
        except:
            raise Exception(print_exc())
    
    def _all_lats_lons(self):
        """Get 2 arrays including all latitudes and all longitudes of all 
        points included in this collection
        
        .. note::
        
            Existing points specifying the regime (i.e. lower left / top 
            right corner) are not considered here
            
        """
        lats, lons = [], []
        for id, p in self.points.items():
            if not any([id == x for x in ["ll","tr"]]):
                lats.append(p.latitude)
                lons.append(p.longitude)
        return asarray(lats), asarray(lons)
    
    @property
    def magnitude(self):
        """Returns dimension (in km) of area covered by this setup"""
        return (self.tr - self.ll).norm
        
    def set_borders_from_points(self, extend_km=1, to_square=True):
        """Set range of setup (lower left and upper right coordinates) 
        considering all points in this collection
        
        :param float extend_km: extend range from the outermost points by 
            this number in km
        :param float to_square (True): extend the shorter base side to the 
            size of the longer one (quadratic range)
        """
        lats, lons= self._all_lats_lons()
        if not len(lats) > 0:
            #print "Borders could not be initiated, no objects found..."
            return False
        
        lat_ll, lon_ll, lat_tr , lon_tr = (nanmin(lats), nanmin(lons),
                                           nanmax(lats), nanmax(lons))
                                           
        pll, ptr = GeoPoint(lat_ll, lon_ll, 0.0), GeoPoint(lat_tr, lon_tr, 0.0)
        
        if to_square:
            v = ptr - pll
            add = (abs(v.dx) - abs(v.dy)) / 2
            if add > 0: #E/W extend (dx) is smaller than N/S extend (dy)
                pll = pll.offset(azimuth = 180, dist_hor = add)
                ptr = ptr.offset(azimuth = 0, dist_hor = add)
            else:
                pll = pll.offset(azimuth = 270, dist_hor = -add)
                ptr = ptr.offset(azimuth = 90, dist_hor = -add)
                 
        self.set_geo_point("ll", pll.offset(azimuth=-135, 
                                            dist_hor=float(extend_km),
                                            name="ll"))
        
        self.set_geo_point("tr", ptr.offset(azimuth=45,
                                            dist_hor=float(extend_km),
                                            name="tr"))
        return True
   
    def create_map(self, *args, **kwargs):     
        """Create a Basemap object for this regime"""
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create map: "
                              "Basemap library is not available")
        
        if not isinstance(self.topo_data, TopoData):
            self.load_topo_data()
        if not "projection" in kwargs and self.magnitude < 150:
            kwargs["projection"] = "lcc"
        if not "llcrnrlon" in kwargs:
            kwargs["llcrnrlat"] = self.lat_ll
            kwargs["llcrnrlon"] = self.lon_ll
            kwargs["urcrnrlat"] = self.lat_tr
            kwargs["urcrnrlon"] = self.lon_tr
            kwargs["lat_0"], kwargs["lon_0"] = self.center_coordinates
       
        m = Map(*args, **kwargs)
        m.set_topo_data(self.topo_data)
        return m
    
    def points_close(self, p, radius=None):
        """Finds all GeoPoints which are within a certain radius around another
        point
        
        :param GeoPoint p: the actual point for which the search is performed
        :param float radius (None): radius (in km) specifying considered range
            around point (is set to 10% of the magnitude of this setup if 
            unspecified)
        :returns:
            - list of point string IDs which are within the specified radius
                around input point
        """
        if radius == None:
            radius = self.magnitude * .1
        ids = []
        for pt in list(self.points.values()):
            if not pt is p and (pt - p).magnitude < radius:
                ids.append(pt.name)
        print(("Found %d points within radius of %.1f km of point %s" 
                                                %(len(ids), radius, p.name)))
        return ids
        
    def plot_2d(self, draw_all_points=True, draw_all_vectors=True,
                draw_topo=True, draw_coastline=True, draw_mapscale=True,
                draw_legend=True, *args, **kwargs):
        """Draw overview map of the current setup
        
        :param bool draw_all_points (True): if true, all points are included
        :param bool draw_all_vectors (True): if true, all vectors (with anchor) 
            are included
        :param bool draw_topo (True): include topography into map
        :param bool draw_coastline (True): include coastline into map
        :param bool draw_mapscale (True): insert map scale
        :param bool draw_legend (True): insert a (draggable) legend
        :param *args: additional non-keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        :param **kwargs: additional keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        :return: 
            - :class:`geonum.mapping.Map` object
            
        """
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create overview map: Basemap module "
                              "is not available")
        if not "ax" in kwargs:
            #fig, ax = subplots(1,1)
            fig = figure(figsize=(10,8))
            ax = fig.add_axes([0.12,0.15,0.8,0.8])
            kwargs["ax"] = ax
        m = self.create_map(*args, **kwargs)
        if draw_coastline:
            m.drawcoastlines()
        if draw_topo:
            m.draw_topo(insert_colorbar=True)
            m.draw_topo_contour()        
        m.draw_coordinates()
        if draw_mapscale:
            m.draw_mapscale_auto()
        p_close_count = 0
        if draw_all_points:
            dist = self.magnitude * .05
            for pt in list(self.points.values()):
                if not any([pt.name == x for x in ["ll", "tr"]]):
                    m.draw_geo_point_2d(pt)
                    ang = -45
                    num_close = len(self.points_close(pt))
                    if num_close > 0:
                        step = 360. / (4 * num_close)
                        ang = ang - step * p_close_count
                        p_close_count += 1
                    m.write_point_name_2d(pt, dist, ang)
        
        #create some color indices for colormap
        nums = [int(255.0 / k) for k in range(1, len(self.vectors)+3)]
        if draw_all_vectors:
            for i, vec in enumerate(self.vectors.values()):
                m.draw_geo_vector_2d(vec, 
                                     ls="-",
                                     c=self.cmap(nums[i]),
                                     label=vec.name)
        if draw_legend:
            try:
                m.legend()
            except:
                warn("Failed to draw legend in GeoSetup...")
                
        return m
                
    def plot_3d(self, draw_all_points=True, draw_all_vectors=True, 
                cmap_topo="Oranges", contour_color="#708090", 
                contour_lw=0.2, contour_antialiased=True, *args, **kwargs):
        """Create a 3D overview map of the current setup
        
        Parameters
        ----------
        draw_all_points : bool
            if True, all current GeoPoint objects are plotted, defaults to 
            True
        draw_all_vectors : bool
            if True, all current GeoVector3D objects are plotted, defaults to
            True
        cmap : str
            string ID of the colormap used to plot the topographic data, 
            defaults to "Oranges"
        contour_color : str
            string specifying color of contour lines colors of contour lines 
            (default: "#708090")
        contour_lw : 
            width of drawn contour lines, defaults to 0.5, use 0 if you do not 
            want contour lines inserted
        contour_antialiased : bool
            apply antialiasing to surface plot of topography, defaults to False
        *args : 
            additional non-keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        **kwargs : 
            additional keyword parameters (passed to `basemap 
            <http://matplotlib.org/basemap/api/basemap_api.html#mpl
            _toolkits.basemap.Basemap>`_)
        
        Returns
        -------
        Map
            plotted 3D basemap
        """
        if not BASEMAP_AVAILABLE:
            raise ImportError("Cannot create overview map: Basemap module "
                              "is not available.")
        m = self.create_map(*args, **kwargs)
        m.draw_topo_3d(cmap=cmap_topo, contour_color=contour_color, 
                       contour_lw=contour_lw, 
                       contour_antialiased=contour_antialiased)
        if draw_all_points:
            zr = self.topo_data.alt_range * 0.05
            alts = []
            for name, pt in self.points.items():
                if not any([name == x for x in ["ll", "tr"]]):
                    try:
                        add_alt = 0 #in m
                        for alt in alts:
                            if abs(alt - pt.altitude) < zr:
                                add_alt = 3 * zr
                                print("Add " + str(add_alt))
                                
                        pt.plot_3d(m, add_name = True, dz_text = zr + add_alt)
                        alts.append(pt.altitude)
    
                    except Exception as e:
                        warn("Point %s could not be drawn: %s"
                             %(pt.name, repr(e)))
                        pass
        if draw_all_vectors:
            nums = [int(255.0 / k) for k in range(1, len(self.vectors)+3)]
            for i, vec in enumerate(self.vectors.values()):
                try:
                    m.draw_geo_vector_3d(vec, label=vec.name,
                                         c=self.cmap(nums[i]),
                                         ls="-",
                                         **kwargs)
                except Exception as e:
                    warn("Vector %s could not be drawn: %s"
                         %(vec.name, repr(e)))
                    pass
                
        return m
    
def show_coordinate(geo_point=None, lat_pt=None, lon_pt=None, extend_km=10.0, 
                    *args, **kwargs):
    """Draw overview map for a given point
    
    Parameters
    ----------
    geo_point : GeoPoint
        Geographical location around which overview map is drawn
    lat_pt : float
        Latitude of geographical location around which overview map is 
        drawn (only considered if :attr:`geo_point` is invalid)
    lon_pt : float
        Longitude of geographical location around which overview map is 
        drawn (only considered if :attr:`geo_point` is invalid)
    extend_km : float
        map extend in km around considered geolocation
    *args :
        non-keyword arguments passed to :func:`plot_2d` of the 
        :class:`GeoSetup` instance that is created in order to draw the map
    
    Returns
    -------
    Map
        instance of :class:`geonum.Map`
        
    """
    if not isinstance(geo_point, GeoPoint):
        try: 
            geo_point = GeoPoint(lat=lat_pt, lon=lon_pt)
        except:
            raise TypeError("Invalid input, please provide information "
                            "about location of GeoPoint")
    stp = GeoSetup(points=[geo_point])
    stp.set_borders_from_points(extend_km=extend_km)
    m = stp.plot_2d(*args, **kwargs)
    return m 


# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

def check_requirements():
    from warnings import warn
    
    BASEMAP_AVAILABLE = True
    CV2_AVAILABLE = True
    NETCDF_AVAILABLE = True
    LATLON_AVAILABLE = True
    SRTM_AVAILABLE = True
    
    try:
        from LatLon23 import LatLon
    except:
        try:
            from LatLon import LatLon
        except:
            warn('Neither LatLon23 nor LatLon are available. Many basic features '
                 'will not be available (e.g. objects GeoPoint or GeoVector ')
            LATLON_AVAILABLE = False
    try:
        import srtm
    except:
        SRTM_AVAILABLE = False
    try:
        from mpl_toolkits.basemap import Basemap
    except:
        warn('Plotting of maps etc. is deactivated, please install Basemap')
        BASEMAP_AVAILABLE = False
    
    try:
        from cv2 import pyrUp
    except:
        CV2_AVAILABLE = False
    try:
        from netCDF4 import Dataset
    except:
        NETCDF_AVAILABLE = False
        
    return (LATLON_AVAILABLE, 
            SRTM_AVAILABLE, 
            BASEMAP_AVAILABLE, 
            CV2_AVAILABLE, 
            NETCDF_AVAILABLE)
        
from os.path import abspath, dirname, join
from pkg_resources import get_distribution

(LATLON_AVAILABLE, 
 SRTM_AVAILABLE, 
 BASEMAP_AVAILABLE, 
 CV2_AVAILABLE, 
 NETCDF_AVAILABLE) = check_requirements()

__dir__ = abspath(dirname(__file__))
__version__ = get_distribution('geonum').version

_LIBDIR = __dir__ #from older version

LOCAL_TOPO_PATH = join(_LIBDIR, "local_topo_data")

from . import helpers
from . import atmosphere
from .topodata import TopoData, TopoDataAccess

if LATLON_AVAILABLE:
    from .base import GeoPoint, GeoVector3D
    from .geosetup import GeoSetup
    from . import base
    from . import geosetup

if LATLON_AVAILABLE:
    from .processing import LineOnGrid, ElevationProfile  
    from . import processing

if BASEMAP_AVAILABLE:
    from .mapping import Map
    from . import mapping
    



# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Access and handling of topography data

.. note::

    The SRTM dataset has no global coverage, it is therefore recommended
    to download one of the Etopo1 data set files as a backup if SRTM data
    cannot be accessed.
    
"""

from numpy import argmin, mod, ceil, log2, poly1d, polyfit,\
    arange, linspace, empty, nan, asarray, nanmax, nanmin, isnan

from os.path import basename, exists, join, dirname, normpath
from os import listdir
from warnings import warn
# python 2 and 3 support see e.g.
# http://python-future.org/compatible_idioms.html#metaclasses
from six import with_metaclass

from abc import ABCMeta, abstractmethod
 
from geonum import (NETCDF_AVAILABLE, CV2_AVAILABLE, LATLON_AVAILABLE, 
                    SRTM_AVAILABLE)
if LATLON_AVAILABLE:
    try:
        from LatLon23 import LatLon
    except:
        from LatLon import LatLon
        
class TopoFile(with_metaclass(ABCMeta, object)):
    """Abstract base class for topgraphy file implementations"""
    local_path = None
    topo_id = None
    
    @abstractmethod
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Declaration of data access method 
        It is obligatory to implement this method into inhereted classes
        """
        pass
    
    def _prep_borders(self, lat0, lon0, lat1, lon1):
        """Sort by longitudes and determines LL and TR coordinates
        
        :param float lat0: latitude of first point
        :param float lon0: longitude of first point
        :param float lat1: latitude of first point
        :param float lon1: longitude of first point
        :return: 
            - float, smallest latitude 
            - float, smallest longitude
            - float, largest latitude
            - float, largest longitude
        """
        lats, lons = asarray([lat0, lat1]), asarray([lon0, lon1])
        return nanmin(lats), nanmin(lons), nanmax(lats), nanmax(lons)
        
    def _init_lons_lats(self, lats_all, lons_all, lat0, lon0, lat1=None, 
                        lon1=None):
        """Get all latitudes and longitudes on a topodata grid 
        
        :param array lats_all: numpy array with all latitudes of the topo data 
            file        
        :param array lons_all: numpy array with all longitudes of the topo data 
            file
        :param float lat0: latitude of point of interest
        :param float lon0: longitude of point of interest
        :param float lat1: latitude of an (optional) second point
        :param float lon1: longitude of an (optional) second point
        :return: 
            - ndarray, latitudes
            - ndarray, longitudes
        """
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0
        if lon0 > lon1:
            lon0, lon1 = lon1, lon0
            lat0, lat1 = lat1, lat0
        #print lat0, lon0, lat1, lon1     
        #closest indices
        idx_lons = [argmin(abs(lons_all - lon0)), argmin(abs(lons_all - lon1))]
        idx_lats = [argmin(abs(lats_all - lat0)), argmin(abs(lats_all - lat1))]
        #Make sure that the retrieved indices actually INCLUDE the input ranges
        if idx_lons[0] == 0 and lons_all[0] > lon0:
            warn("Error: Lon0 smaller than range covered by file, using first"
                " available index in topodata..")
            lon0 = lons_all[0]
            idx_lons[0] = 0
        elif lons_all[idx_lons[0]] > lon0:
            idx_lons[0] -= 1
        if idx_lons[1] == len(lons_all) - 1 and lons_all[-1] < lon1:
            warn("Error: Lon1 larger than range covered by file, using last"
                " available index in topodata..")
            lon1 = lons_all[-1]
            idx_lons[1] = len(lons_all) - 1
        elif lons_all[idx_lons[1]] < lon1:
            idx_lons[1] += 1
        if idx_lats[0] == 0 and lats_all[0] > lat0:
            warn("Error: Lat0 smaller than range covered by file, using first"
                " available index in topodata..")
            lat0 = lats_all[0]
            idx_lats[0] = 0
        elif lats_all[idx_lats[0]] > lat0:
            idx_lats[0] -= 1
        if idx_lats[1] == len(lats_all) - 1 and lats_all[-1] < lat1:
            warn("Error: Lat1 larger than range covered by file, using last"
                " available index in topodata..")
            lat1 = lats_all[-1]
            idx_lats[1] = len(lats_all) - 1
        elif lats_all[idx_lats[1]] < lat1:
            idx_lats[1] += 1
        #make sure that no odd array lengths occur
        if not (idx_lats[1] - idx_lats[0] + 1) %2 == 0:
            #try append index at the end
            if not idx_lats[1] == len(lats_all) - 1:
                idx_lats[1] += 1
            elif not idx_lats[0] == 0:
                idx_lats[0] -= 1
            else:
                raise ValueError("Fatal error, odd length of latitude array")
        if not (idx_lons[1] - idx_lons[0] + 1) %2 == 0:
            #try append index at the end
            if not idx_lons[1] == len(lons_all) - 1:
                idx_lons[1] += 1
            elif not idx_lons[0] == 0:
                idx_lons[0] -= 1
            else:
                raise ValueError("Fatal error, odd length of longitude array")
        if idx_lats[0] > idx_lats[1]:
            return (lats_all[idx_lats[1] : idx_lats[0] + 1], 
                    lons_all[idx_lons[0] : idx_lons[1] + 1],
                    idx_lats, idx_lons)
        else:
            return (lats_all[idx_lats[0] : idx_lats[1] + 1],
                    lons_all[idx_lons[0] : idx_lons[1] + 1],
                    idx_lats, idx_lons)
        
class Etopo1Access(TopoFile):
    """A class representing netCDF4 data access of Etopo1 data
    
    See `here <https://github.com/jgliss/geonum#supported-etopo1-files>`_ for
    instructions on the data access.
    
    """
    def __init__(self, local_path=None, file_name=None):
        """Class initialisation
        
        :param str local_path: directory where Etopo data files are stored
        :param str file_name: file name of etopo data file
        
        """
        if not NETCDF_AVAILABLE:
            raise ModuleNotFoundError("Etopo1Access class cannot be initiated. "
                                      "Please install netCDF4 library first")
        from netCDF4 import Dataset
        self.loader = Dataset
        self.topo_id = "etopo1"
        
        self.local_path = local_path
        self.file_name = file_name

        self.supported_topo_files = ["ETOPO1_Ice_g_gmt4.grd",
                                     "ETOPO1_Bed_g_gmt4.grd"]
        #checks input and if applicable adds to database
        self._check_topo_path(local_path) 
        if local_path is None or not exists(local_path):
            from geonum import LOCAL_TOPO_PATH
            self.local_path = LOCAL_TOPO_PATH
        
        # make sure the file is supported
        if not file_name in self.supported_topo_files:
            # this function only returns valid file names, else None
            self._search_topo_file() #searches current local_path for valid file
        if not exists(self.file_path):
            self.search_topo_file_database()
        
        # check if file exists
        if not exists(join(self.local_path, self.file_name)):
            raise TopoAccessError("File %s could not be found in local "
                "topo directory: %s" %(self.file_name, self.local_path))
    
    @property
    def file_path(self):
        """Return full file path of current topography file"""
        return join(self.local_path, self.file_name)
    
    def _check_topo_path(self, path):
        """Check if path exists and if it is already included in database"""
        if path is None or not exists(path):
            return 
        if not path in self._get_all_local_topo_paths():
            from geonum import LOCAL_TOPO_PATH
            with open(join(LOCAL_TOPO_PATH, "LOCAL_TOPO_PATHS.txt"), "a") as f:
                f.write("\n" + path  + "\n")
                print(("Adding new default local topo data path to "
                    "file LOCAL_TOPO_DATA.txt: %s" %path))
            f.close()
        
    def _get_all_local_topo_paths(self):
        """Get all search paths for topography files"""
        from geonum import LOCAL_TOPO_PATH
        paths = [LOCAL_TOPO_PATH]
        with open(join(LOCAL_TOPO_PATH, "LOCAL_TOPO_PATHS.txt"), "r") as f:
            lines = f.readlines()
            for line in lines:
                p = line.split("\n")[0]
                if exists(p):
                    paths.append(normpath(p))
        f.close()
        return paths
        
    def _check_local_topo_path(self, path):
        """Checks if local topo path exists and is part of database"""
        if path is None or not exists(path):
            return False
        
    def _search_topo_file(self, path=None):
        """Checks if a valid etopo data file can be found in local folder
        
        Searches in ``self.local_path`` for any of the file names specified 
        in ``supported_topo_files``
        
        """
        if path is None:
            path = self.local_path
        print(("Searching valid topo file in folder: %s" %path))
        fnames = listdir(path)
        for name in fnames:
            if name in self.supported_topo_files:
                self.file_name = name
                self.local_path = path
                print(("Found match, setting current filepath: %s" 
                %self.file_path))
                return True
        return False
        
    def _find_supported_files(self):
        """Look for all supported files in ``self.local_path```and return list"""
        files = listdir(self.local_path)
        lst = []
        for name in files:
            if name in self.supported_topo_files:
                lst.append(name)
        return lst
    
    def search_topo_file_database(self):
        """Checks if a valid topo file can be found in database"""
        all_paths = self._get_all_local_topo_paths()
        for path in all_paths:
            if self._search_topo_file(path):
                return True
        return False
        
    def set_file_location(self, full_path):
        """Set the full file path of a topography data file
        
        :param str full_path: file path with topo data
        """
        if not exists(full_path):
            raise TopoAccessError("Input file location %s does not exist"
                                                                %full_path)
        if not basename(full_path) in self.supported_topo_files:
            raise TopoAccessError("Invalid topography data file, please use "
                "one of the supported files from the Etopo1 data set\n%s" 
                %self.supported_topo_files)
        self.local_path = dirname(full_path)
        self.file_name = basename(full_path)
    
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Retrieve data from topography file 
        
        :param float lon0: start longitude for data extraction
        :param float lat0: start latitude for data extraction
        :param float lon1: stop longitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        :param float lat1: stop latitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        
        :returns:
            - ndarray (2D), map containing retrieved elevations
            - ndarray, array containing corresponding latitudes
            - ndarray, array containing corresponding longitudes
        
            
        """
#==============================================================================
#         print ("Trying topo data access Etopo1 within borders (Lon0 | Lat0), : "
#             "(Lon1 | Lat1): (%s | %s), (%s | %s)" %(lat0, lon0, lat1, lon1))
#==============================================================================
    
        etopo1 = self.loader(self.file_path)

        lons = etopo1.variables["x"][:]
        lats = etopo1.variables["y"][:]
        
        lats, lons, idx_lats, idx_lons = self._init_lons_lats(lats, lons, lat0,
                                                              lon0, lat1, lon1)

        vals = asarray(etopo1.variables["z"][idx_lats[0] : idx_lats[1] + 1,
                                             idx_lons[0] : idx_lons[1] + 1], 
                                             dtype = float)
        etopo1.close()

        return TopoData(lats, lons, vals, data_id=self.topo_id)
         
class SRTMAccess(TopoFile):
    """Class for SRTM topographic data access (uses 
    `SRTM <https://pypi.python.org/pypi/SRTM.py/0.3.1>`_ library for 
    access).
    
    .. note::
    
        :mod:`SRTM` downloads the topo data from 
        `here <https://www.ngdc.noaa.gov/mgg/global/global.html>`_
        `the official web page <http://dds.cr.usgs.gov/srtm/version2_1/>`_ and
        stores a copy of the unzipped data files in the current home directory, 
        i.e., either::
            import os
            os.path.join(os.environ["HOMEPATH"], ".cache\\srtm")
            
        or::
        
            os.path.join(os.environ["HOME"], ".cache\\srtm")
        
        Whenever data access is requested, the SRTM library checks if the file
        already exists on the local machine and if not downloads it online. The
        online access is rather slow, so do not be surprised, if things take a
        while when accessing a specific location for the first time.   
          
    """
    def __init__(self):
        """Class initialisation"""
        if not SRTM_AVAILABLE:
            raise ModuleNotFoundError("SRTM access class cannot be initiated. "
                                      "Please install srtm.py library first")
        import srtm
        self.loader = srtm
        self.topo_id = "srtm"
        
    def coordinate_covered(self, access_obj, lat, lon):
        """Checks if SRTM data is available for input coordinate
        
        :param GeoElevationData access_obj: data access object from 
            :mod:`srtm` module (can be created calling ``srtm.get_data()``)
        :param float lat: latitude of point
        :param float lon: longitude of point
        :returns: - bool
        """
        if access_obj.get_file_name(lat, lon) is None:
            return False
        return True

        
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Data access wrapper for SRTM module
        
        Retrieves SRTM data in the speciefied input range.
        
        :param float lon0: start longitude for data extraction
        :param float lat0: start latitude for data extraction
        :param float lon1: stop longitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        :param float lat1: stop latitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        
        :returns:
            - ndarray (2D), map containing retrieved elevations
            - ndarray, array containing corresponding latitudes
            - ndarray, array containing corresponding longitudes
        
        """
        print("Retrieving SRTM data (this might take a while) ... ")
        # create GeoElevationData object for data access
        dat = self.loader.get_data()
        # check if second input point is specified and set equal first point if
        # not
        if any([x is None for x in [lat1, lon1]]):
            lat1, lon1 = lat0, lon0
        # Check if first point is covered by dataset
        if not self.coordinate_covered(dat, lat0, lon0):
            raise SRTMNotCoveredError("Point not covered by SRTM (Lat|Lon): "
                                                    "(%s | %s)" %(lat0, lon0))
        # check if second point is covered by dataset
        if not self.coordinate_covered(dat, lat1, lon1):
            raise SRTMNotCoveredError("Destination point not covered by SRTM ("
                "Lat | Lon): (%s | %s)" %(lat0, lon0))
        # prepare borders of covered lon / lat regime
        lat_ll, lon_ll, lat_tr,lon_tr = self._prep_borders(lat0, lon0,
                                                           lat1, lon1)
        # get SRTM file for lower left corner of regime
        f_ll = dat.get_file(lat_ll, lon_ll)
        # get SRTM file for top right corner of regime
        f_tr = dat.get_file(lat_tr, lon_tr)
        # create array of longitude values for regime
        lons_all = linspace(f_ll.longitude, f_tr.longitude + 1,
                            f_ll.square_side)
        # create array of latitude values for regime
        lats_all = linspace(f_ll.latitude, f_tr.latitude + 1,
                            f_ll.square_side)
        #prepare coordinates
        lats, lons, _, _= self._init_lons_lats(lats_all, lons_all,
                                               lat0, lon0, lat1, lon1)
        # Init data array 
        vals = empty((len(lats), len(lons))) * nan
        #loop over all coordinates and try access the elevation data
        for i in range(len(lats)):
            for j in range(len(lons)):
                #print "Lat: %s, Lon: %s" % (lats[i], lons[j])
                vals[i, j] = dat.get_elevation(lats[i], lons[j])
    
        return TopoData(lats, lons, vals, data_id=self.topo_id)
        
class TopoDataAccess(object):
    """Factory class for accessing topographic data
    
    This is a high level access class which handles the access of topography
    data. It is, for instance, implemented within :class:`geonum.base.GeoPoint` 
    or :class:`geonum.geosetup.GeoSetup` classes. 
    
    Default access mode is SRTM. 
    """
    def __init__(self, mode="srtm", local_path=None):
        """Class initialisation
        
        :param str mode: choose between one of the supported data
            access types (currently 2: etopo1, srtm)
        :param str local_path: local path to etopo data (only relevant for 
            etopo1 mode)
        """
        #mode and data access variables
        self.mode = None
        self.modes = ["srtm", "etopo1"]
        self.local_path = None
        self.topo_file = SRTMAccess()
        
        self.set_mode(mode, local_path)
            
    def init_default_mode(self):
        """Set default mode (e.g. called when object is initiated with etopo1
        but without a file specification)
        """
        self.mode = "srtm"
        self.topo_file = SRTMAccess()
    
    def set_mode(self, mode="srtm", local_path=None):
        """Change the current topography mode 
        
        :param str mode: the new mode
        :param str local_path (None): option to update the current local_path 
            variable (i.e. the location of etopo1 grid files)
            
        :raises InvalidTopoMode: if input mode is not suppored
        :raises TopoAccessError: if mode "etopo1" and corresponding data 
            file cannot be found at ``self.local_path``
        """
        if local_path is not None: 
            if not exists(local_path):
                raise ValueError("Failed to set local topography path, path "
                                 "does not exist: %s" %local_path)
            self.local_path = local_path
        
        
        if not mode in self.modes:
            raise InvalidTopoMode("Mode %s not supported...\nSupported modes:"
                "%s\nCurrent mode: %s " %(mode, self.modes, self.mode))
        if mode == "etopo1":
            if self.local_path is None or not exists(self.local_path):
                raise IOError("Could not change to mode Etopo1, local path to "
                    "topography dataset not set, please download")
            tf = Etopo1Access(self.local_path)
            if tf.file_path is not None:
                self.mode = mode
                self.topo_file = tf
                return
            self.init_default_mode()
            raise TopoAccessError("Could not access Etopo1 data on local "
                                            "machine, check local_path...")
        
        self.init_default_mode() #which means SRTM
    
    def get_data(self, lat0, lon0, lat1=None, lon1=None):
        """Retrieve data from topography file and, if applicable, increase the 
        resolution by upscaling (Gaussian pyramide approach) and interpolation
        
        Data access wrapper for SRTM module
        
        Retrieves SRTM data in the speciefied input range.
        
        :param float lon0: start longitude for data extraction
        :param float lat0: start latitude for data extraction
        :param float lon1: stop longitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
        :param float lat1: stop latitude for data extraction (default: None). 
            If None only data around lon0, lat0 will be extracted
                
        :returns: :class:`TopoData` object containing the data
        :raises: :class:`TopoAccessError`  if access fails
        
        """
        topo_data = None
        try:
            lat0, lon0 = float(lat0), float(lon0)
        except Exception as e:
            raise TopoAccessError(repr(e))
        try:
            topo_data = self.topo_file.get_data(lat0, lon0, lat1, lon1)
        except Exception as e:
            print(("Topo retrieval failed using mode " + str(self.mode) +
                                ", try extracting data from Etopo1 dataset"))
            msg = repr(e)
            try:
                self.set_mode("etopo1")
            except (InvalidTopoMode, TopoAccessError) as err:
                msg = "Error loading topodata: " + msg + "\n" + repr(err)
                raise TopoAccessError(msg)

            try:
                topo_data = self.topo_file.get_data(lat0, lon0, lat1, lon1)
                
            except Exception as err:
                msg = "Error loading topodata: %s\n%s" %(msg, repr(err))
                raise TopoAccessError(msg)
        try:
            print("Important remarks while accessing topodata: %s" %msg)
        except:
            pass
        return topo_data
        
class TopoData(object):
    """Class representing topography data as 2D numpy grid specified by corner
    coordinates (lon0, lat0, lon1, lat1), normally created by 
    :class:`TopoDataAccess`.
    
    .. note::
    
        This object does not provide topography data access. For data access, 
        please use :class:`TopoDataAccess` which returns :class:`TopoData` 
        objects. This object is intended for storage and post processing of
        topographic data 
        
    """
    def __init__(self, lats, lons, data, data_id="", repl_nan_minval=True):
        """Class initialisation
        
        :param ndarray lats: numpy array with latitude coordinates of the 
            topographic dataset
        :param ndarray lons: numpy array with longitude coordinates of the 
            topographic dataset
        :param ndarray data: 2D numpy array containing elevation values
        :param str data_id: string identification of this data set
        :param bool repl_nan_minval: pixels containing NaN values are 
            replaced with the minimum altitude in the range
        """
        self.data_id = data_id #: ID of topodata file

        self.lats = lats #asarray(lats)
        self.lons = lons #asarray(lons)
        if repl_nan_minval:
            data[isnan(data)] = nanmin(data)
        
        self.data = data
        
    @property
    def lon0(self):
        """Returns first longitude (i.e. ``self.lons[0]``)"""
        return self.lons[0]
        
    @property
    def lat0(self):
        """Returns first latitude (i.e. ``self.lats[0]``)"""
        return self.lats[0]
        
    @property
    def lon1(self):
        """Returns last longitude (i.e. ``self.lons[-1]``)"""
        return self.lons[-1]
        
    @property
    def lat1(self):
        """Returns last latitude (i.e. ``self.lats[-1]``)"""
        return self.lats[-1]
        
    @property
    def max(self):
        """Returns maximum altitude of topo data"""
        return nanmax(self.data)
        
    @property
    def min(self):
        """Returns minimum altitude of topo data"""
        return nanmin(self.data)
        
    @property
    def alt_range(self):
        """Returns covered altitude range"""
        return self.max - self.min
        
    @property
    def resolution(self):
        """Returns tuple (lat, lon) of the current grid resolution in km
        
        .. note::
        
            The resolution is determined at the center of this grid
        """
        if not LATLON_AVAILABLE:
            raise ModuleNotFoundError('Feature disabled: Neither LatLon nor '
                                      'LatLon23 are installed')
        x_lon, x_lat = int(len(self.lons) / 2), int(len(self.lats) / 2)
        p0 = LatLon(self.lats[x_lat], self.lons[x_lon])
        r_lon = (p0 - LatLon(self.lats[x_lat], self.lons[x_lon + 1])).magnitude
        r_lat = (p0 - LatLon(self.lats[x_lat + 1], self.lons[x_lon])).magnitude
        return (r_lat, r_lon)
        
    def increase_grid_resolution(self, res=0.2, polyorder=2):
        """Gaussian pyramide based upscaling of topographic grid
        
        This function checks the current topographic resolution in the center
        of the grid. Based on this, an upsacling factor is determined and the
        :func:`cv2.pyrUp` is used to increase the resolution using
        interpolation. Note, that this does not increase the actual resolution
        of the topographic data grid.
        
        :param float res: desired grid resolution in km (default: 0.2)
        :param int polyorder: order of polynomial used for interpolation
            (default: 2)
        :returns: 
            - :class:`TopoData`, new object with desired grid resolution
            
        .. note::
        
            This functionality is only available, if :mod:`cv2` is installed
            
            
        """
        if not CV2_AVAILABLE or not LATLON_AVAILABLE:
            raise ModuleNotFoundError('Feature disabled: Require opencv and '
                                      'LatLon (or LatLon23) library to change '
                                      'grid resolution ')
        from cv2 import pyrUp
        lons = self.lons
        lats = self.lats
        vals = self.data
        if not all(mod(x, 2) == 0 for x in [len(lons), len(lats)]):
            print ("Fatal error, odd array size detected, no upscaling possible."
                " Return current data dict")
            return False
            
        c_lon = len(lons) / 2 - 1 #center longitude index
        c_lat = len(lats) / 2 - 1 #center latitude index
        p1 = LatLon(lats[c_lat], lons[c_lon]) 
        p2 = LatLon(lats[c_lat + 1], lons[c_lon + 1])
        dist = (p2 - p1).magnitude #distance between 2 points on grid
        res_fac = dist / res #factor for increasing the spatial resolution
        if res_fac <= 1:
            print(("No interpolation of topodata necessary: topo raw data "
                "already has desired resolution...\nCurrent resolution: %s" 
                + str(dist) + "km\nDesired resolution: %s km" %(dist, res)))
            return self
        fac = int(ceil(log2(res_fac))) #the corresponding up-factor for the scale space
        print(("Increasing spatial topography resolution by factor %s" %(2**fac)))
        for k in range(fac):
            vals = pyrUp(vals)
        p_lons = poly1d(polyfit(arange(len(lons)), lons, polyorder))
        p_lats = poly1d(polyfit(arange(len(lats)), lats, polyorder))
        lons_new = p_lons(linspace(0, len(lons) - 1, vals.shape[1]))
        lats_new = p_lats(linspace(0, len(lats) - 1, vals.shape[0]))
        return TopoData(lats_new, lons_new, vals, self.data_id + "_interp")

    @property
    def delta_lon(self):
        """Longitude range of data (in decimal degrees)"""
        return self.lons[-1] - self.lons[0]
        
    @property 
    def delta_lat(self):
        """Latitude range of data (in decimal degrees)"""
        return self.lats[-1] - self.lats[0]
        
    @property
    def center_coordinates(self):
        """Tuple (lat, lon) with center coordinates of data"""
        return (self.lats[0] + self.delta_lat / 2.,
                self.lons[0] + self.delta_lon / 2.)
    
    def plot_2d(self, ax=None):
        """Plot 2D basemap of topodata"""
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1,
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo(insert_colorbar=True)
        return m
        
    def plot_3d(self, ax=None):
        """Creates 3D surface plot of data
        
        :param Axes3D ax (None): 3D axes object
        :return:
            - :class:`geonum.mapping.Map` object
            
        """
        from geonum.mapping import Map
        latc, lonc = self.center_coordinates
        m = Map(self.lon0, self.lat0, self.lon1, self.lat1, 
                projection="merc", lat_0=latc, lon_0=lonc, ax=ax)
        m.set_topo_data(self)
        m.draw_topo_3d()
        return m
        
    def includes_coordinate(self, lat=None, lon=None):
        """Checks if input coordinates are covered by this dataset"""
        if not (self.lat0 <= lat <= self.lat1 and 
                self.lon0 <= lon <= self.lon1):
            print("Input coordinates out of topo data range...")
            return False
        return True
    
    def get_altitude(self, lat, lon):
        """Get altitude value at input position
        
        :param float lat: latitude of point
        :param float lon: longitude of point
        """
        if not self.includes_coordinate(lat, lon):
            raise ValueError("Input values out of range...")
        idx_lon = argmin(abs(self.lons - lon))
        idx_lat = argmin(abs(self.lats - lat))
        dat = self.data[idx_lat, idx_lon]
        if isnan(dat):
            raise ValueError("Invalid value encountered in topodata...")
        return dat
        
    def __call__(self, lat, lon):
        """Get altitude value at input position"""
        try:
            return self.get_altitude(lat, lon)
        except ValueError as e:
            raise e
        
class SRTMNotCoveredError(Exception):
    pass

class TopoAccessError(Exception):
    pass

class InvalidTopoMode(Exception):
    pass

def delete_all_local_srtm_files():
    """Deletes all locally stored SRTM files"""
    raise NotImplementedError
    #from srtm.main import FileHandler
    #fh = FileHandler()
#==============================================================================
#     dir = fh.get_srtm_dir()
#     filelist = [ f for f in os.listdir(dir) if f.endswith(".bak") ]
#==============================================================================
if __name__ == "__main__":
    import matplotlib.pyplot as plt
    topopath = r'D:/Dropbox/Python27/jgliss/modules_priv/pygeomapping/topodata/'
    
    etopo_cell = "ETOPO1_Ice_c_gmt4.grd"
    etopo_grid = "ETOPO1_Ice_g_gmt4.grd"
    
    lon0, lat0 = 14.85, 37.7
    lon1, lat1 = 15.2, 37.85
    cell = Etopo1Access(topopath, etopo_cell)
    grid = Etopo1Access(topopath, etopo_grid)
    
    srtm_access = SRTMAccess()
    
    d1 = cell.get_data(lat0, lon0, lat1, lon1)
    d2 = grid.get_data(lat0, lon0, lat1, lon1)
    d3 = srtm_access.get_data(lat0, lon0, lat1, lon1)    
    
    fig, ax = plt.subplots(1,3,figsize=(18,6))
    ax[0].imshow(d1[0])
    ax[1].imshow(d2[0])
    ax[2].imshow(d3[0])
    
    access = TopoDataAccess()
    td = access.get_data(lat0, lon0, lat1, lon1)


# -*- coding: utf-8 -*-
"""Test environment for base.py module."""
from __future__ import (absolute_import, division)
from geonum.base import GeoPoint, GeoVector3D


def test_GeoPoint():
    """Test basic arithmetic operations on GeoPoints."""
    assert(GeoPoint(1, 1))


def test_GeoVector3D():
    """Test basic arithmetic operations on GeoVectors."""
    assert(GeoVector3D(1, 1))



# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
This modules includes the two base objects of geonum:
    
    1. :class:`GeoPoint` specifies a coordinate geographical coordinate 
    (lon, lat, altitude)
    
    2. :class:`GeoVector3D` specifies a vector (e.g. connection between 2 
    points)
        
"""
try:
    from LatLon23 import LatLon, GeoVector
except:
    from LatLon import LatLon, GeoVector
    
from numpy import radians, cos, sin, degrees, sqrt,tan, isnan, arctan2,\
    asarray, nanmin, nanmax

from warnings import warn
from geonum.helpers import isnum
from geonum.topodata import TopoDataAccess, TopoData

class GeoPoint(LatLon):
    """The Geopoint object represents a point on an ellipsoid (e.g. earth) 
    including elevation information. It inherits from :class:`LatLon` object 
    from the :mod:`LatLon` library.
    """
    def __init__(self, lat=0.0, lon=0.0, altitude=None, name="n/d",
                 topo_access_mode="srtm", topo_path=None, topo_data=None,
                 auto_topo_access=True):
                        
        """Class initialisation
        
        :param float lat: latitude of point (decimal degrees)
        :param float lon: longitude of point (decimal degrees)
        :param float altitude: elevation (above surface) of point in m
        :param str name: name (ID) of this point
        :param str topo_access_mode: string specifying the current access 
            mode for topographic data (in v1, choose between "srtm" or 
            "etopo1")
        :param str topo_path: specify path where etopo1 data files are 
            stored
        :param TopoData topo_data: you can assign an existing topo dataset 
            to this point (can save time, e.g. for altitude access)
        """
        lat, lon = float(lat), float(lon)
        if not all([isnum(x) for x in [lat, lon]]):
            raise ValueError("Invalid input for lat, lon in GeoPoint, got "
                             "%s, %s (%s, %s) need non-NaN numeric" %(lat, lon,
                                                              type(lat),
                                                              type(lon)))
        #super(GeoPoint, self).__init__(lon, lat, name)
        LatLon.__init__(self, lat, lon, name)
        self.altitude = altitude #altitude in m
        self.altitude_err = 0.0
        
        self._topo_access = TopoDataAccess(mode=topo_access_mode,
                                           local_path=topo_path)
        self.topo_data = None 
        self.set_topo_data(topo_data)
        
        if auto_topo_access and (self.altitude is None or\
                                 isnan(self.altitude)):
            self.get_altitude()
    
    @property
    def local_topo_path(self):
        """Returns current etopo1 data access path (str)"""
        return self._topo_access.local_path    
    
    """PROCESSING"""  
    def offset(self, azimuth, dist_hor, dist_vert=0.0, ellipse="WGS84", 
               **kwargs):
        """Returns new GeoPoint at offset position
    
        :param float azimuth: azimuth direction angle (in decimal degrees, e.g.
            90 for E direction, -90 or 270 for west)
        :param float dist_hor: horizontal offset to this point in km 
        :param float dist_vert: vertical offset to this point in m (positive
            if point is higher, negative, if it is lower)
        :param float ellipse: geodetic system (ellipsoid), default is 
            "WGS84", i.e. `World Geodetic System <https://confluence.qps.
            nl/pages/view page.action?pageId=29855173>`_)
        :param **kwargs: additional keyword arguments passed to init of new
            :class:`GeoPoint` object (e.g. name = "blablub")
        :return: - new :class:`GeoPoint` at offset position
        """
        p = LatLon(self.lat.decimal_degree, self.lon.decimal_degree)
        p1 = p.offset(azimuth, dist_hor, ellipse)
        return GeoPoint(p1.lat.decimal_degree, p1.lon.decimal_degree,
                        self.altitude + dist_vert, **kwargs)
    
    def set_topo_data(self, topo_data):
        """Assign topographic data to this point
        
        If input is valid, the topographic data set will be stored in 
        class attribute topo_data.
        
        :param TopoData topo_data: the topo data set        
        """
        if not isinstance(topo_data, TopoData):
            return
        elif not topo_data.includes_coordinate(self.latitude, self.longitude):
            raise TypeError("Could not assign topodata to GeoPoint %s: "
             "topodata does not cover the coordinates" %self.name)
        self.topo_data = topo_data
    
    def range_borders(self, *points):
        """Get geographical borders (lower left, top right) of this point
        
        Additional points can be included as non keyword arguments. The
        range borders are then determined under consideration of all
        specified points. If no additional points are specified,
        the range corners will be set corresponding to 1km diagonal 
        distance to this point.
        
        :param *points: additional :class:`GeoPoint` objects to be 
            considered
        :returns:
            - :class:`GeoPoint`, lower left corner of regime
            - :class:`GeoPoint`, top right corner of regime
        
        """
        lats, lons = [self.latitude], [self.longitude]
        #retrieve all latitudes and longitudes
        for p in points:
            if isinstance(p, GeoPoint):
                lats.append(p.latitude)
                lons.append(p.longitude)
        lats, lons = asarray(lats), asarray(lons)
        if not len(lats) > 0:
            #print "Borders could not be initiated, no objects found..."
            return False
        lat_ll, lon_ll, lat_tr , lon_tr = (nanmin(lats), nanmin(lons), 
                                           nanmax(lats), nanmax(lons))
        pll, ptr = GeoPoint(lat_ll, lon_ll, 0.0), GeoPoint(lat_tr, lon_tr, 0.0)
        extend_km = (pll - ptr).magnitude * 0.1
        if extend_km == 0:
            extend_km = 1
        ll = pll.offset(azimuth=-135, dist_hor=float(extend_km), name="ll")
        tr = ptr.offset(azimuth=45, dist_hor=float(extend_km), name="tr")
        return  ll, tr        
        
    def get_topo_data(self, geo_point=None, azimuth=None, dist_hor=None,
                      lon1=None, lat1=None):
        """Retrieve topographic data grid spanned by this and another point
        
        The second coordinate can be specified in several different ways 
        dependent on which of the input arguments are used. First checks if the
        currently loaded topo data (``self.topo_data``) covers the range 
        spanned by the 2 points. If this is the case, the loaded data will be 
        used and nothing reloaded. If not, data will attempted to be loaded 
        from the current :class:`TopoDataAccess` object within the lon lat 
        range covered by the 2 points (set using :func:`range_borders`)
        
        :param GeoPoint geo_point: another geo_point,topo data will be 
            retrieved between this point and destination point
        :param float azimuth: azimuth angle of heading in decimal degrees 
            (to be used with dist_hor)
        :param float dist_hor: horizontal distance from this point in km 
            (to be used with azimuth)
        :param float lon1: longitude of destination point, topo data will be 
            retrieved between this point and destination point 
            (to be used with lat1)
        :param float lat1: latitude of destination point, topo data will be 
            retrieved between this point and destination point 
            (to be used with lon1)
        :returns: 
            - :class:`TopoData`, the topographic data
            - :class:`GeoPoint`, the second coordinate
          
        The following input combinations work (and are preferentially 
        processed in the specified list order if multiple input is given):
        
            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1
        
        """
        pf = self
        if isinstance(geo_point, GeoPoint):
            pf = geo_point
        elif all(isinstance(x, (int, float)) for x in [dist_hor, azimuth]):
            pf = self.offset(azimuth, dist_hor)
        elif all(isinstance(x, (int, float)) for x in [lon1, lat1]):
            pf = GeoPoint(lat1, lon1)
        
        ll, tr = self.range_borders(self, pf)
        lat0, lon0 = ll.latitude, ll.longitude
        lat1, lon1 = tr.latitude, tr.longitude
        if not (self.check_topo(lat1, lon1) and self.check_topo(lat0, lon0)):
            print(("Topo data not avaialable in geo point %s, loading topodata"
                        %self.name))
            self._load_topo_data(lat0, lon0, lat1, lon1)
        else:
            print ("Topo data avaialable in geo point %s")
        return self.topo_data, pf
    
    def check_topo(self, lat1=None, lon1=None):
        """Check if topography is available between this point and another
        
        :param float lat1: latitude of end point, if None, only the 
            coordinates of this object will be considered
        :param float lon1: longitude of end point, if None, only the 
            coordinates of this object will be considered
        :returns: - bool, True if topo data is loaded, False if not
        """
        if any([x == None for x in [lat1, lon1]]):
            lat1, lon1 = self.latitude, self.longitude
        if self.topo_data is None:
            return False
        if not (self.topo_data.includes_coordinate(self.latitude, 
                                                   self.longitude) 
            and self.topo_data.includes_coordinate(lat1, lon1)):
            return False
        return True
        
    def get_elevation_profile(self, geo_point=None, azimuth=None, 
                              dist_hor=None, lon1=None, lat1=None, 
                              resolution=5.):
        """Estimates the elevation profile for a given viewing direction up
        to a given distance from the coordinates of this point. For input 
        possibilities see docs of :func:`get_topo_data`.
        
        :param GeoPoint geo_point: another geo_point,topo data will be 
            retrieved between this point and destination point
        :param float azimuth: azimuth angle of heading in decimal degrees 
            (to be used with dist_hor)
        :param float dist_hor: horizontal distance from this point in km 
            (to be used with azimuth)
        :param float lon1: longitude of destination point, topo data will 
            be retrieved between this point and destination point (to be 
            used with lat1)
        :param float lat1: latitude of destination point, topo data will be 
            retrieved between this point and destination point 
            (to be used with lon1)
        :param float resolution: desired topo grid resolution in m. 
            1D interpolation of the elevation profile is performed is 
            applicable.
        :returns: 
            - :class:`ElevationProfile`, the profile object
          
        The following input combinations work (and are preferentially processed
        in the specified list order if multiple input is given):
        
            1. specify endpoint using geo_point
            #. specify endpoint using azimuth and dist_hor
            #. specify endpoint using lon1 and lat1
            
        
        """
        from geonum.processing import ElevationProfile
        data, pf = self.get_topo_data(geo_point, azimuth, dist_hor, lon1, 
                                      lat1)
        return ElevationProfile(data, self, pf, resolution=resolution)
        
    def get_altitude(self):
        """Estimate the altitude (+/- uncertainty) of this point 
        
        The estimatation is done by retrieving a 2x2 grid of topography
        data enclosing this point. The altitude value is estimated using the 
        topo data tile closest to the coordinates of this point. The uncertainty
        is estimated conservatively using min/max difference of data in the 2x2
        grid.
        
        :returns: 
            - altitude (0 if retrieval failed)
            - altitude error (99999999999 if retrieval failes)
        :rtype:
            - int, float
        
        """
#==============================================================================
#         print ("Try retrieving altitude for point %s (lat, lon): %s, %s" 
#                                 %(self.name, self.latitude, self.longitude))
#==============================================================================
        try:
            #the data access class already catches access failure and in case,
            #tries to  access etopo1 data
            data = self._topo_access.get_data(self.latitude,
                                              self.longitude)
            z = data(self.latitude, self.longitude)# / 1000.
            z_err = (data.data.max() - data.data.min()) #/ 1000.'
#            print "Retrieved altitude: %s +/- %s m" %(z, z_err)
        except:
            warn("Altitude could not be retrieved...setting altitude to 0.0 m")
            z, z_err = 0.0, 99999999999.9
    
        self.altitude, self.altitude_err = z, z_err
        
        return z, z_err
    
    """HELPERS, IO stuff, etc"""
    def update_topo_access(self, mode, local_path=None):
        """Update topo access mode
        
        :param str mode: valid access mode (e.g. "srtm", "etopo1")        
        :param str local_path (None): path for etopo1 access (if None, uses 
            the one which is currently set in ``self._topo_access``)
        """
        if local_path is None:
            local_path = self._topo_access.local_path
        self._topo_access = TopoDataAccess(mode=mode, 
                                           local_path=local_path)
        
    @property
    def topo_access_mode(self):
        """Return current topography access mode"""
        return self._topo_access.mode
        
    @property
    def longitude(self):
        """Returns longitude coordinate as decimal number"""
        return self.lon.decimal_degree

    @property
    def latitude(self):
        """Returns longitude coordinate as decimal number"""
        return self.lat.decimal_degree
        
    """Plotting etc..."""
    def plot_2d(self, map, add_name=False, dist_text=0.5, angle_text=-45, 
                **kwargs):
        """Plot this point into existing 2D basemap
                
        :param map basemap: Basemap object (drawn in an Axes3D object)
        :param bool add_name: add the name of this GeoPoint in the map
        :param float dist_text: distance of text annotation from point
        :param float angle_text: angle of text displacement
        :param kwargs: additional keyword arguments passed to matplotlib 
            plot function
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_2d(self, **kwargs)
        if add_name:
            map.write_point_name_2d(self, dist_text, angle_text)
                                                    
    def plot_3d(self, map, add_name=False, dz_text=0.0, **kwargs):
        """Plot this point into existing 3D basemap
                
        :param map basemap: Basemap object (drawn in an Axes3D object)
        :param bool add_name: add the name of this GeoPoint in the map
        :param float dz_text: altitude offset of text(to point)
        :param kwargs: additional keyword arguments passed to matplotlib 
            plot function
        """
        if not "marker" in kwargs:
            kwargs["marker"] = "^"
        if not any([x in kwargs for x in ["c", "color"]]):
            kwargs["c"] = "lime"
        map.draw_geo_point_3d(self, **kwargs)
        x0, y0 = list(map(self.lon.decimal_degree, self.lat.decimal_degree))
        if add_name and self.name is not None:
            try:
                fs = kwargs["fontSize"]
            except:
                fs = 12
            #zt=self.altitude*1000. + dz_text
            zt = self.altitude + dz_text
            map.draw_text_3d(self.longitude, self.latitude, zt, self.name,
                             color="k", fontsize=fs)
            
    """PRIVATE METHODS"""  
    def _load_topo_data(self, lat0, lon0, lat1, lon1):
        """Load and update topo data
        
        :param float lon0: start longitude
        :param float lat0: start latitude
        :param float lon1: stop longitude 
        :param float lat1: stop latitude
        :return:
            - :class:`TopoData` object with loaded data (it is also stored in
                ``self.topo_data``)
        """
        print("Trying to load topography data in GeoPoint obj..")
        self.topo_data = self._topo_access.get_data(lat0, lon0, lat1, lon1)
        return self.topo_data
        
    def _sub_geo_vector_2d(self, other):
        """Subtract another geo vector
        
        Called when subtracting a GeoVector object from self (adapted from
        :class:`LatLon` object,  only return type was changed from LatLon 
        object to GeoPoint object)
        
        :param (GeoVector, GeoVector3D) other: vector to be subtracted
        :return: :class:`GeoPoint` at difference position
        """
        heading, distance = other()
        heading = (heading + 180) % 360 # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude)
        
        return p.offset(heading, distance)
    
    def _add_geo_vector_2d(self, other):
        """Subtract another geo vector
        
        Add a geo vector (adapted from :class:`LatLon` object,  only return 
        type was changed from LatLon object to GeoPoint object)
        
        :param GeoVector other: Geovector added to this 
        
        """
        azimuth, dist_hor = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, 0.0)
        
    def _sub_geo_vector_3d(self, other):
        """Called when subtracting a GeoVector3D object from self"""
        azimuth, dist_hor, dz = other()
        azimuth = (azimuth + 180) % 360 # Flip heading
        p = GeoPoint(self.lat, self.lon, self.altitude) # Copy current position
        return p.offset(azimuth, dist_hor, -dz) # Offset position by GeoVector
    
    def _add_geo_vector_3d(self, other):
        """Add a 3d geo vector object
        
        :param GeoVector3D other: Geovector which is added to self
        """
        azimuth, dist_hor, dz = other()
        p = GeoPoint(self.lat, self.lon, self.altitude)
        return p.offset(azimuth, dist_hor, dz)
        
    def _sub_latlon(self, other):
        """Called when subtracting a LatLon object from this point"""
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        name=str(other.name + "->" + self.name)
        return GeoVector3D(dz=0.0, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)
    
    def _sub_geo_point(self, other):
        """Called when subtracting a LatLon object from self"""
        inv = self._pyproj_inv(other)
        heading = inv['heading_reverse']
        distance = inv['distance']
        name = str(other.name + "->" + self.name)
        dz = self.altitude - other.altitude
        return GeoVector3D(dz=dz, azimuth=heading, dist_hor=distance,
                           anchor=other, name=name)
        
    """Redfined magic methods initially implemented for :class:`LatLon` object
    """    
    def __eq__(self, other):
        """Checks equality of this point with another
        
        :param other: another :class:`GeoPoint` object
        """
        if (self.lat == other.lat and self.lon == other.lon 
            and self.altitude == other.altitude):
            return True
        return False
        
    def __add__(self, other):
        """Add a geo vector to this point
        
        :param (GeoVector, GeoVector3D) other: a vector
        """
        object_operator = {'GeoVector'      :   self._add_geo_vector_2d,
                           'GeoVector3D'    :   self._add_geo_vector_3d}
        return object_operator[other.type()](other)

    def __sub__(self, other):
        """Subtraction"""
        object_operator = {'GeoVector'      :   self._sub_geo_vector_2d,
                           'GeoVector3D'    :   self._sub_geo_vector_3d,
                           'LatLon'         :   self._sub_latlon,
                           'GeoPoint'       :   self._sub_geo_point}
        return object_operator[other.type()](other)
    
    def __str__(self):
        """String formatting"""
        return ("GeoPoint %s\nLat: %s, Lon: %s, Alt: %s m\n" 
                %(self.name, self.latitude, self.longitude, self.altitude))
    
    def __repr__(self):
        """Obj. representation"""
        return ("%s, %s, Alt. %s m" 
                %(self.lat.__repr__(), self.lon.__repr__(), self.altitude))
    
    def __complex__(self):
        """Complex representation of lat and lon"""
        return self.complex()
    
    def type(self):
        """Object type identifier
        
        :returns: str, the string identifier        
        """
        return 'GeoPoint'


class GeoVector3D(GeoVector):
    """A 3 dimensional geo vector object 
    
    3D vector representation for geo numerical calculations, intuitive 
    usage, i.e.::
    
        from geonum import GeoPoint, GeoVector3D
        p = GeoPoint(10.0, 15.0, name = "random_point") #lat, lon
        v = GeoVector3D(dx = 15, dy = 100, dz = 300) #dx, dy in km, dz in m
        
        new_point = p + v # GeoPoint object
        
    
    .. note:: 
    
        This class inherits and makes use of the functionality of 
        :class:`GeoVector` objects of the :mod:`LatLon` library. Methods and 
        attributes are partly the same and partly overwritten
        (note that it is not initiated as :class:`GeoVector`)
 
    """
            
    def __init__(self, dx=None, dy=None, dz=None, azimuth=None, dist_hor=None, 
                 elevation=None, anchor=None, name="n/d"):
        """Class initialisation
        
        :param float dx: longitudinal length in km
        :param float dy: latitudinal length in km
        :param float dz: altitude length in m
        :param float azimuth: azimuth angle to North direction (corresponds to 
            heading_initial variable in :class:`GeoVector` object)
        :param float dist_hor: horizontal displacement of object in km
        :param float elevation: elevation angle in decimal degrees 
            (0 corresponds to horizon, 90, to zenith)
        
        For initiation use one of the following to input combinations:
        
            1. dx, dy, dz
            #. dz, azimuth, dist_hor
            #. azimuth, dist_hor, elevation
            
        Multiple input combinations will be processed in the preferred order
        as given in the list
        
        """
        self.name = name
        if any(x == None for x in [dx, dy]): # If only initial_heading and distance are given
            theta_rad = radians(self._angle_or_heading(azimuth))
            self.dx = dist_hor * cos(theta_rad)
            self.dy = dist_hor * sin(theta_rad)
        elif azimuth == None and dist_hor == None: # If only dx and dy are given
            self.dx = dx
            self.dy = dy
        else:
            raise NameError
        #Check input for altitude difference dz
        if dz is None or isnan(dz): #invalid for dz directly
            if elevation is not None and -90 <= elevation <= 90: #check if instead elevation is valid, then set dz
                #tan elev = dz/dist_hor
                dz = tan(radians(elevation)) * sqrt(self.dx**2 + self.dy**2)\
                                                                        * 1000
            else: #both dz input and elevation are invalid, set dz=0
                dz = 0.0
        self.dz = dz
        # dictionary with private attributes
        self._priv_attr= {"anchor" : None}
        # call setter for private attribute anchor (this ensures that input 
        # attr anchor is of right type
        try:
            self.set_anchor(anchor)
        except:
            pass
    
    @property
    def azimuth(self):
        """Azimuth angle in dec degrees 
        
        Horizontal orientation measured from N direction
        """
        return self._geom_hor()[0]

    @property
    def elevation(self):
        """Elevation angle in decimal degrees
        
        Measured from horizon (i.e. 0 -> horizon, 90 -> zenith)
        """
        return 90 - self.polar_angle    
    
    @property
    def polar_angle(self):
        """Polar angle in decimal degrees
        
        Returns the polar angle in decimal degrees (measured from zenith), 
        i.e.::
        
            90 - self.elevation
            
        """
        return degrees(arctan2(self.dist_hor, self.dz / 1000.))
    
    @property    
    def magnitude(self):
        """Return magnitude of vector (length)"""
        return sqrt(self.dist_hor**2 + (self.dz / 1000.)**2)

    @property
    def norm(self):
        """Return magnitude of vector (length)"""
        return self.magnitude
        
    @property
    def dist_hor(self):
        """Horizontal distance spanned by this vector"""
        return self._geom_hor()[1]
    
    @property
    def anchor(self):
        """Getter method for private attribute anchor"""
        return self._priv_attr["anchor"]
        
    @anchor.setter
    def anchor(self, value):
        """Setter method of private property anchor
        
        See :func:`set_anchor` for input / output specs
        """     
        self.set_anchor(value)
    
    def set_anchor(self, geo_point):
        """Set anchor of this vector
        
        :param GeoPoint geo_point: anchor point of this vector
        :raises TypeError: if input is not :class:`GeoPoint` object        
        """
        if not isinstance(geo_point, GeoPoint):
            raise TypeError("Could not set anchor: Invalid input type")
        self._priv_attr["anchor"] = geo_point
           
    def intersect_hor(self, other):
        """Determine the horizontal intersection of this vector with other 
        input vector
        
        :param GeoVector3D: other vector
        
        .. note::
        
            Only works if anchor is defined
        """
        if not all([isinstance(x,GeoPoint) for x in [self.anchor, other.anchor]]):
            raise ValueError("Intersection can not be determined, anchor "
                " of one of the vectors not set..")
        v = other.anchor - self.anchor
        v.dz = 0.0
        other_az = radians((other.azimuth + 360) % 360)
        self_az = radians((self.azimuth + 360) % 360)
        dy = (v.dx - tan(other_az) * v.dy) / (tan(self_az) - tan(other_az))
        dx = tan(self_az) * dy
        return GeoVector3D(dx, dy, dz=0.0)
        
    def _geom_hor(self):
        """Returns horizontal heading and horizontal magnitude"""
        return (degrees(arctan2(self.dx, self.dy)),
                sqrt(self.dx**2 + self.dy**2))
        
    """Plotting etc..."""
    def plot(self, map, add_anchor=False, **kwargs):
        """Plot this vector into existing basemap
        
        :param Map map: map object
        :param bool add_anchor: If true, the anchor point is plotted as well
        :param kwargs: additional keyword arguments
        
        .. note::
        
            1. The basemap needs to be set up with Axes3D
            #. ``self.anchor`` must be set with a :class:`Geopoint` instance
            
        """
        map.draw_geo_vector_3d(self, **kwargs)
        if add_anchor:
            self.anchor.plot(map, add_name=True, dz_text=self.dz*.1)
        
    """Redifining magic methods from base class :class:`GeoVector` object for
    3D calcs
    """
    def __call__(self):
        """Call function
        
        :returns:
            - float azimuth angle
            - float, horizontal length in km
            - float, vertical length in m
        """
        azimuth, dist_hor = self._geom_hor()
        return azimuth, dist_hor, self.dz
    
    
        
    def __neg__(self):
        """Returns negative of this vector"""
        return GeoVector3D(-self.dx, -self.dy, -self.dz)
    
    def __add__(self, other):
        """Add another geo vector
        
        
        :param GeoVector3D other: another geo vector  
        """
        return GeoVector3D(self.dx + other.dx, self.dy + other.dy,
                           self.dz + other.dz)

    def __str__(self):
        """String representation"""
        return ('GeoVector3D %s\nAzimuth: %s, Elevation: %s, Magnitude: %s m\n'
                %(self.name,self.azimuth, self.elevation, self.magnitude))
            
    
    def __repr__(self):
        return ('Az %s, Elev %s, Mag. %s' 
                %(self.azimuth, self.elevation, self.magnitude))
    
    def type(self):
        """Returns object type"""
        return 'GeoVector3D'



# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Processing module of geonum library
"""
from numpy import radians, cos, sin, arctan, pi, argmin, linspace,\
    tan, ceil, hypot, vstack, nanmin, nanmax, where, sign, diff,\
    logical_and, arange, nan, gradient, rad2deg
    #gradient
from warnings import warn
from scipy.ndimage import map_coordinates
from scipy.interpolate import interp1d

from matplotlib.pyplot import subplots, tight_layout

from geonum.base import GeoPoint

class ElevationProfile(object):
    """Class for calculating elevation profiles
    
    The profile is calculated between two geo point objects using a 
    provided topographic data grid which must cover the range spanned by 
    both points
    
    Parameters
    ----------
    topo_data : TopoData 
        topography data object
    observer : :obj:`GeoPoint` or :obj:`LatLon`
        starting point of profile 
    endpoint : :obj:`GeoPoint` or :obj:`LatLon`
        stop point of profile 
    interpolate : bool
        if True, the profile is interpolated to a certain horizontal resolution
    resolution : float
        desired grid resolution in m for interpolation. Interpolation is 
        performed if :attr:`interpolate` is True and if the actual 
        resolution of the topo data is smaller than input, else, nothing is done
    itp_type : str
        interpolation type (e.g. "linear", "cubic")
    """
# =============================================================================
#     def __new__(cls, topo_data, observer, endpoint, interpolate=True, 
#                 resolution=5.0, itp_type="linear"):
#         """These objects strictly can only be created with the right input, 
#         
#         For input specs see __init__ method
#         """
#         if not isinstance(topo_data, TopoData):
#             raise ValueError("ElevationProfile instance could not be created\n"
#                 "Wrong input type: topo_data, need TopoData object...")
#         for p in [observer, endpoint]:
#             if not any([p.type() == x for x in ["LatLon", "GeoPoint"]]):
#                 raise ValueError ("ElevationProfile instance could not be "
#                     "created\nWrong input type: observer/endpoint, need GeoPoint "
#                     "or LatLon object...")
#             elif not topo_data.includes_coordinate(p.lat.decimal_degree,
#                                                    p.lon.decimal_degree):
#                 raise ValueError ("ElevationProfile instance could not be "
#                     "created\nWrong input point not covered by input topodata")
#         
#         return super(ElevationProfile, cls).__new__(cls, topo_data, observer,
#                                                     endpoint, interpolate,
#                                                     resolution, 
#                                                     itp_type)
# =============================================================================
        
    def __init__(self, topo_data, observer, endpoint, interpolate=True, 
                 resolution=5.0, itp_type="linear"):
     
        self.topo_data = topo_data
        self.observer = observer #: coordinate of observer (start of profile)
        self.endpoint = endpoint #: coordinate of endpoint of profile
        
        self._observer_topogrid = None #: closest to observer on topo data grid
        self._endpoint_topogrid = None #: closest to endpoint in topo data grid
        
        # In the following parameters the results will be stored
        self.line = None
        self.profile = None
        self.dists = None

        try:
            self.det_profile(interpolate, resolution, itp_type)
        except Exception as e:
            warn("Failed to compute elevation profile. Error msg: %s"
                 %repr(e))
        
    @property
    def dist_hor(self):
        """Returns the horizontal distance between the 2 points"""
        return (self._endpoint_topogrid - self._observer_topogrid).dist_hor
        
    
    @property
    def azimuth(self):
        """Returns the azimuth angle of the profile"""
        return (self._endpoint_topogrid - self._observer_topogrid).azimuth
        
    @property
    def resolution(self):
        """Get profile x (distance) resolution (averaged from distance array)
        
        .. note::

            Only works if profile was already determined
            
        """
        return abs((self.dists[1:] - self.dists[:-1]).mean())
    
    @property
    def gradient(self):
        """Return gradient of profile
        
        Uses numpy function ``gradient``
        """
        return gradient(self.profile)
    
    @property
    def slope(self):
        """Returns slope of profile 
        
        Determines dx and dy arrays and returns slope vector::
        
            slope = dy / dx = gradient(self.profile) / 
                (gradient(self.dists) * 1000.0)
            
        """
        return gradient(self.profile) / (gradient(self.dists) * 1000)
    
    def slope_angles(self, decimal_degrees=True):
        """Returns slope angle of profile (in each sample point)
        
        :param bool decimal_degrees: rad or degrees (default True)
        """
        a = tan(self.slope)
        if decimal_degrees:
            a = rad2deg(a)
        return a
        
    def slope_angle(self, dist):
        """Returns slope angle of profile at input distance
        
        :param float dist: distance in km
        """
        idx = argmin(abs(self.dists - dist))
        return self.slope_angles()[idx]
        
    def det_profile(self, interpolate=True, resolution=5.0, itp_type="linear"):
        """Determines the elevation profile
        
        Searches the closest tiles in the topo data grid for both observer and
        endpoint and based on these two points the elevation profile is
        extracted using a :class:`LineOnGrid` object (which extracts altitudes 
        from the topo data along the connection vector of the 2 points using 
        2D spline intepolation)
        
        Parameters
        ----------
        interpolate : bool
            if True, the profile is interpolated to a certain horizontal resolution
        resolution : float
            desired grid resolution in m for interpolation. Interpolation is 
            performed if :attr:`interpolate` is True and if the actual 
            resolution of the topo data is smaller than input, else, nothing is done
        itp_type : str
            interpolation type (e.g. "linear", "cubic")
        
        Returns
        -------
                
        """
        data = self.topo_data
        idx_lon_0 = argmin(abs(data.lons - self.observer.lon.decimal_degree))
        idx_lat_0 = argmin(abs(data.lats - self.observer.lat.decimal_degree))
        idx_lon_1 = argmin(abs(data.lons - self.endpoint.lon.decimal_degree))
        idx_lat_1 = argmin(abs(data.lats - self.endpoint.lat.decimal_degree))
        
        self.line = l = LineOnGrid(idx_lon_0, idx_lat_0, 
                                   idx_lon_1, idx_lat_1)
        
        z = l.get_line_profile(data.data)
        self._observer_topogrid = GeoPoint(data.lats[idx_lat_0],\
                                data.lons[idx_lon_0], topo_data = data)
        self._endpoint_topogrid = GeoPoint(data.lats[idx_lat_1],\
                                data.lons[idx_lon_1], topo_data = data)
        dists = linspace(0, self.dist_hor, l.length + 1)
        if interpolate:
            try:
                res0 = (dists[1] - dists[0]) * 1000
                fac = int(ceil(res0 / resolution))
                if fac > 1:
                    fz = interp1d(dists, z, kind=itp_type)
                    dists = linspace(0, self.dist_hor, l.length * fac)
                    z = fz(dists)
            except Exception as e:
                warn("Failed to perform interpolation of retrieved elevation "
                     "profile. Error msg: %s" %repr(e))
        self.dists = dists
        self.profile = z
     
    def get_altitudes_view_dir(self, elev_angle, view_above_topo_m=1.5):
        """Get vector containing altitudes for a viewing direction 
        
        The viewing direction is specified by the azimuth angle of the 
        connection vector between observer and endpoint, the elevation angle
        needs to be specified via the input parameters, and the start point
        (first elevation value) corresponds to the altitude at the observer
        position plus an offset in m which can be specified.
        
        :param float elev_angle: elevation angle of viewing direction
        :param float view_above_topo_m (1.5): altitude offset of start point 
            in m
        :return:
            - vector with altitude values (same length as ``self.profile``)
            
        """
        return (1000 * tan(radians(elev_angle)) * self.dists + 
                self.profile[0] + view_above_topo_m)
    
    def find_horizon_elev(self, elev_start=0.0, elev_stop=60.0, step_deg=0.1,
                          **kwargs):
        """Find first elevation angle which does not intersect with topo
        
        :param float elev_start: start search elevation angle
        :param float elev_stop: stop search elevation angle
        :param float step_deg: angle step for search (coarser is faster)
        :param **kwargs: additional keyword agruments passed to 
            :func:`get_first_intersection`
        """
        elevs = arange(elev_start, elev_stop + step_deg, step_deg)
        elev_sects = []
        dists_sects = []
        for elev in elevs:
            (dist, 
             dist_err, 
             intersect, 
             view_elevations, 
             _) = self.get_first_intersection(elev, **kwargs)
            if dist is None:
                return elev, elev_sects, dists_sects
            else:
                dists_sects.append(dist), elev_sects.append(elev)
        raise Exception("Unexpected exception..")
        
    def get_first_intersection(self, elev_angle, view_above_topo_m=1.5,
                               min_dist=None, local_tolerance=3, plot=False):
                                
        """Finds first intersection of a viewing direction with topography
        
        Start point of the viewing vector is the observer position (or more
        accurately, the center position of the closest topography tile in 
        the topography data set). The relative altitude of the start point 
        with respect to the topography altitude at the observer position 
        can be controlled on input (arg ``view_above_topo_m``) as well as 
        the elevation angle of the viewing direction. The azimuth angle 
        corresponds to the profile azimuth (i.e. azimuth between observer 
        position and endpoint of this profile).
        
        The signal analysed for finding the intersection is a vector 
        containing relative elevation positions of the viewing direction 
        with respect to the profile altitude::
        
            diff_signal = self.get_altitudes_view_dir - self.profile
            
        The first intersection of the viewing direction with the topography
        is identified by the first zero crossing of this ``diff_signal``.
        
        :param float elev_angle: elevation angle of viewing direction (in 
            decimal degrees)
        :param int view_above_topo_m: altitude offset of start point 
            (``observer``) in m (default: 1.5)
        :param float min_dist: minimum distance (in km) of first 
            considered intersection with topography from observer. If None, 
            use 1% of distance between ``observer`` and ``endpoint``.
        :param float local_tolerance: tolerance factor to estimate 
            distance uncertainty based on topo grid resolution  (default: 3)
        :param bool plot: creates a plot of the profile including the
            intersection information

        
        """
         
        if min_dist == None:
            min_dist = self.dist_hor*.01
        max_diff = self.resolution * 1000 
        view_elevations = self.get_altitudes_view_dir(elev_angle,
                                                      view_above_topo_m)
        
        #: determine the difference signal
        diff_signal = view_elevations - self.profile 
        #: First condition: consider only points in certain distance from observer
        cond1 = self.dists > min_dist
        #: Second condition: consider only "close to zero" points (this might be
        #: redundant, I'll leave it for now because it works)
        cond2 = abs(diff_signal) < max_diff 
        
        #: create array with all distances matching the 2 conditions
        dists_0 = self.dists[cond1 * cond2] 
        dist, dist_err, intersect = None, None, None
        #: relax condition 2 if nothing was found 
        if not len(dists_0) > 0:
            max_diff = self.resolution * 1000 * local_tolerance
            cond2 = abs(diff_signal) < max_diff
            dists_0 = self.dists[cond1 * cond2]
                   
        try:
            #: get all diff vals matching the 2 conditions
            diff_vals = diff_signal[cond1 * cond2]
            #: get the index of the first zero crossing
            first_idx = where(diff(sign(diff_vals)))[0][0]
            
            #: get distance and local tolerance value
            dist, tol = dists_0[first_idx], self.resolution * local_tolerance
            #: make mask to access all distances within tolerance range
            cond4 = logical_and(dist - tol <= dists_0, dist + tol >= dists_0)
            #: estimate distance error from standard deviation of all 
            #: distances within tolerance range
            dist_err = dists_0[cond4].std()
            
            #: create geopoint corresponding to intersection
            intersect = self._observer_topogrid.offset(self.azimuth, dist)
            
            #: set the altitude of this point (retrieved from profile)
            intersect.altitude = self.get_altitude_at_distance(dist) # / 1000.
           
        except IndexError as e:
            print(("No intersections could be detected, err: %s" %repr(e)))
        
        ax = None
        if plot:
            ax = self._plot_intersect_search_result(view_elevations, dist)
            if dist == None:
                dist = nan
            ax.set_title("Azim: %.1f, Elev: %.1f, Intersect @ "
                "dist =%.1f km" %(self.azimuth, elev_angle, dist))
            
        return dist, dist_err, intersect, view_elevations, ax
    
    def _plot_intersect_search_result(self, view_elevations, dist=None):
        ax = self.plot()
        ax.plot(self.dists, view_elevations, label = "Viewing direction")
        try:
            ax.axvline(dist, ls = "--", label = "Intersection")
        except:
            pass
        ax.legend(loc="best", fancybox=True, framealpha=0.4)
        return ax
        
        
    def get_altitude_at_distance(self, dist):
        """Returns altitude at a ceratain distance from observer
        
        :param float dist: horizontal distance from obsever along profile        
        """
        idx = argmin(abs(self.dists - dist))
        return self.profile[idx]
    
    @property
    def start_point(self):
        """Return position of observer"""
        return self.observer
        
    @property
    def min(self):
        """Return minimum altitude in profile"""
        return nanmin(self.profile)
        
    @property
    def max(self):
        """Return maximum altitude in profile"""
        return nanmax(self.profile)
    
    @property
    def alt_range(self):
        """Return altitude range of profile
        """
        return self.max - self.min
        
    def __call__(self, dist):
        """Returns altitude at a certain distance
        
        :param float dist: distance in km
        """
        return self.get_altitude_at_distance(dist)
        
    def plot(self, ax = None):
        """Plot the profile into an axis object
        
        :param ax: matplotlib axis object
        """
        if ax is None:
            fig, ax = subplots(1,1)
        ax.fill_between(self.dists, self.profile, facecolor="#994d00",
                        alpha=0.20)
        ax.set_xlabel("Distance [km]")
        ax.set_ylabel("Altitude [m]")
        ax.set_ylim([self.min - .1 * self.alt_range, 
                     self.max + .1 * self.alt_range])
        return ax
        
class LineOnGrid(object):
    """A class representing a line on a discrete grid"""
    def __init__(self, x0, y0, x1, y1, id=""):
        """Class initialisation
        
        :param int x0: start x coordinate
        :param int y0: start y coordinate
        :param int x1: stop x coordinate
        :param int y1: stop y coordinate
        :param str id: string for identification (optional)
        """
        self.id = id
        self.start  = [x0,y0]
        self.stop   = [x1,y1]
        
        self.profile_coords = None
        
        #self.check_coordinates()
        self._det_normal_vec()
        self.prepare_profile_coordinates()
    
    def check_coordinates(self):
        """Check if coordinates are in the right order and swap, if not"""
        if any([x < 0 for x in self.start]):
            raise ValueError("Invalid value encountered, coords must not be"
                " smaller than 0")
        if any([x < 0 for x in self.stop]):
            raise ValueError("Invalid value encountered, coords must not be"
                " smaller than 0")
        if self.start[0] > self.stop[0]:
            print("X coordinate of Start point is larger than X of stop point")
            print("Start and Stop will be exchanged")
            self.start, self.stop = self.stop, self.start
    
    """Processing functionality"""
    def prepare_profile_coordinates(self):
        """Prepare the evaluation coordinates as stack"""
        length = self.length
        x0, y0 = self.start     
        x1, y1 = self.stop
        x = linspace(x0, x1, length+1)
        y = linspace(y0, y1, length+1)
        self.profile_coords = vstack((y,x))
    
    @property
    def normal_vector(self):
        """Normal vector of line"""
        return self._det_normal_vec()
        
    @property
    def length(self):
        """Determine the length in grid coordinates"""
        del_x, del_y = self._delx_dely()
        return int(round(hypot(del_x, del_y)))
    
    def get_line_profile(self, array):
        """Retrieve line profile of data on a 2D array
        
        :param ndarray array: the data array
        """
        # Extract the values along the line, using cubic interpolation
        zi = map_coordinates(array, self.profile_coords)
        return zi
        
    """Plotting / visualisation etc...
    """
    def plot_line_on_grid(self, array, ax=None, **kwargs):
        """Plot this line on an input array using imshow
        
        :param ndarray array: the data array
        :param **kwargs: additional keyword arguments for imshow
        
        """
        if ax is None:
            fig, ax = subplots(1,1)
        
        im = ax.imshow(array, cmap="gray", interpolation='none', 
                       origin="upper", **kwargs)
        fig.colorbar(im, ax=ax, shrink=0.9)
        ax.plot([self.start[0], self.stop[0]], [self.start[1], self.stop[1]],
                'co-')
            
        ax.set_xlim([0, array.shape[1] - 1])
        ax.set_ylim([array.shape[0] - 1, 0])
        return ax
    
    def plot_line_profile(self, array, ax=None, **kwargs):
        """Plot the line profile 
        
        :param ndarray array: the data array
        :param ax (None): axes object
        :param **kwargs: additional keyword arguments for imshow
        
        """
        if ax is None:
            fig, ax = subplots(1,1)
        p = self.get_line_profile(array)
        ax.set_xlim([0, self.length])
        ax.grid()
        ax.plot(p)
        ax.set_title("Line profile")
        return ax
    
    def plot(self, array, fig_width=12):
        """High level plot function, calls:
        
            1. :func:`plot_line_on_grid`
            #. :func:`plot_line_profile`
            
        and puts them next to each other in a subplot
        
        :param ndarray array: the data array
        :param int fig_width: width of figure in inch
        
        """
        dx, dy = self._delx_dely()
        if dx > dy:
            r = float(dy) / dx
            h = int(fig_width * r * 2.5)
            fig, axes = subplots(2,1, figsize=(fig_width, h))
        else:
            fig, axes=subplots(1,2, figsize = (18, 6))
        self.plot_line_on_grid(array, ax = axes[0])
        self.plot_line_profile(array, ax = axes[1])
        tight_layout()
    
    """'Private' functions"""
    def _delx_dely(self):
        """Returns length of x and y range"""
        return (abs(self.stop[0] - self.start[0]),
                abs(self.stop[1] - self.start[1]))
                        
    def _det_normal_vec(self):
        """Determines the normal vector of the line"""
        delx, dely = self._delx_dely()
        try:
            a = arctan(float(dely) / float(delx))
        except ZeroDivisionError:
            a = pi / 2
        return (sin(a), cos(a))
        
    """Magic methods"""
    def __str__(self):
        """String representation"""
        s = ("LineOnGrid %s\nStart (x,y): %s\nStop (x,y): %s\n" 
                                        %(self.id, self.start, self.stop))
        return s


# -*- coding: utf-8 -*-
#
# Geonum is a Python library for geographical calculations in 3D
# Copyright (C) 2017 Jonas Gliss (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Helper methods for geonum library
"""

from numpy import radians, cos, arcsin, sin, sqrt, array, linspace, hstack,\
    floor, log10, abs, int, isnan
import matplotlib.cm as colormaps
import matplotlib.colors as colors
from matplotlib.pyplot import draw

exponent = lambda num: int(floor(log10(abs(num))))

def isnum(val):
    """Checks if input is number (int or float) or and not nan
    
    :returns: bool, True or False    
    """
    if isinstance(val, (int, float)) and not isnan(val):
        return True
    return False

def rotate_xtick_labels(ax, deg=30, ha="right"):
    """Rotate xtick labels in matplotlib axes object"""
    draw()
    lbls = ax.get_xticklabels()
    lbls = [lbl.get_text() for lbl in lbls]
    ax.set_xticklabels(lbls, rotation = 30, ha = "right")
    draw()
    return ax
    
def haversine_formula(lon0, lat0, lon1, lat1, radius=6371.0):
    """Haversine formula
    
    Approximate horizontal distance between 2 points assuming a spherical 
    earth
    
    :param float lon0: longitude of first point in decimal degrees
    :param float lat0: latitude of first point in decimal degrees
    :param float lon1: longitude of second point in decimal degrees
    :param float lat1: latitude of second point in decimal degrees
    :param float radius (6371.0): average earth radius in km
    """
    hav = lambda d_theta: sin(d_theta / 2.0) ** 2
    
    d_lon = radians(lon1 - lon0)
    d_lat = radians(lat1 - lat0)
    lat0 = radians(lat0)
    lat1 = radians(lat1)
 
    a = hav(d_lat) + cos(lat0) * cos(lat1) * hav(d_lon)
    c = 2 * arcsin(sqrt(a))
 
    return radius * c

def approximate_connection_vector(lon0, lat0, lon1, lat1, len_lat_km=111.20):
    """Returns approximate connection vector between two points
    
    :param float lon0: longitude of first point in decimal degrees
    :param float lat0: latitude of first point in decimal degrees
    :param float lon1: longitude of second point in decimal degrees
    :param float lat1: latitude of second point in decimal degrees
    
    Careful: Only approximative, suited for small distance ( < 100km)
    """    
    
        
    lat = (lat0 + lat1) / 2 * 0.01745
    dx = len_lat_km * cos(lat) * (lon1 - lon0)
    dy = len_lat_km * (lat1 - lat0)
    return array((dx,dy))
    
def shifted_color_map(vmin, vmax, cmap=None):
    """Shift center of a diverging colormap to value 0
    
    .. note::
    
        This method was found `here <http://stackoverflow.com/questions/
        7404116/defining-the-midpoint-of-a-colormap-in-matplotlib>`_ 
        (last access: 17/01/2017). Thanks to `Paul H <http://stackoverflow.com/
        users/1552748/paul-h>`_ who provided it.
    
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and if you want the
    middle of the colormap's dynamic range to be at zero level
    
    :param vmin: lower end of data value range
    :param vmax: upper end of data value range
    :param cmap: colormap (if None, use default cmap: seismic)
    
    :return: 
        - shifted colormap
        
    """
    #midpoint = 1 - abs(im.max())/(abs(im.max()) + abs(im.min()))
    if cmap is None:
        cmap = colormaps.seismic
        
    midpoint = 1 - abs(vmax)/(abs(vmax) + abs(vmin))
    
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = linspace(0, 1, 257)

    # shifted index to match the data
    shift_index = hstack([
        linspace(0.0, midpoint, 128, endpoint=False), 
        linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    #newcmap = colors.LinearSegmentedColormap('shiftedcmap', cdict)
    #register_cmap(cmap=newcmap)

    return colors.LinearSegmentedColormap('shiftedcmap', cdict)


# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...

with open('README.rst') as file:
    readme = file.read()

with open("VERSION.rst") as f:
    version = f.readline()
    f.close()
 
#with open(join("geonum","local_topo_data", "LOCAL_TOPO_PATHS.txt"), 'w'): pass
       
setup(
    name        =   'geonum',
    version     =   version,
    author      =   'Jonas Gliss',
    author_email=   'jg@nilu.no',
    license     =   'GPLv3',
    url         =   'https://github.com/jgliss/geonum',
    package_dir =   {'geonum'     :   'geonum'},
    packages    =   find_packages(exclude=['contrib', 'docs', 'tests*']),
    include_package_data    =   True,  
    package_data=   {'geonum'     :   ['local_topo_data/*.rst',
                                       'local_topo_data/*.txt']},
    install_requires  =   [],
    extras_require={
        'cv2':      ["opencv-python>=2.4.11"],
        'netCDF4':  ["netCDF4>=1.1.8"]},

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.,
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6'
    ],
    
    #dependency_links    =   ["https://github.com/tkrajina/srtm.py/archive/v.0.3.1.zip#egg=srtm"],
    #package_data={'geonum':['suppl/*.dat']},
    description = 'Toolbox for 3D geonumerical calculations',
    long_description = readme,
    
    #requires=['python (>= 2.7)', 'numpy', 'astropy'],
)